# -*- coding: utf-8 -*-

#~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~#
# Std lib
from collections import OrderedDict, Counter, defaultdict
import warnings


# Third party
from loguru import logger
from scipy.stats import mannwhitneyu, ttest_ind, chi2, f_oneway
from scipy.stats.mstats import ks_twosamp
import statsmodels.discrete.discrete_model as dm
from statsmodels.tools.sm_exceptions import ConvergenceWarning
from sklearn.preprocessing import StandardScaler
from sklearn.mixture import GaussianMixture
import numpy as np
import pandas as pd

# Local package
from nanocompore.common import *


def txCompare(
    ref_id,
    ref_pos_list,
    random_state,
    methods=None,
    sequence_context=0,
    min_coverage=20,
    ref=None,
    sequence_context_weights="uniform",
    anova=True,
    logit=False,
    allow_warnings=False):
    logger.debug("TxCompare")

    if sequence_context_weights != "uniform" and sequence_context_weights != "harmonic":
        raise NanocomporeError("Invalid sequence_context_weights (uniform or harmonic)")

    n_lowcov = 0
    tests = set()
    # If we have less than 2 replicates in any condition skip anova and force logit method
    if not all([ len(i)>1 for i in ref_pos_list[0]['data'].values() ]):
        anova=False
        logit=True
    for pos, pos_dict in enumerate(ref_pos_list):
        logger.trace(f"Processing position {pos}")
        # Filter out low coverage positions
        lowcov = False
        for cond_dict in pos_dict["data"].values():
            for sample_val in cond_dict.values():
                if sample_val["coverage"] < min_coverage:
                    lowcov=True
        ref_pos_list[pos]["lowCov"]=lowcov

        # Perform stat tests if not low cov
        if lowcov:
            logger.trace(f"Position {pos} is low coverage, skipping")
            n_lowcov+=1
        else:
            res = dict()
            data = pos_dict['data']
            condition_labels = tuple(data.keys())
            if len(condition_labels) != 2:
                raise NanocomporeError("The %s method only supports two conditions" % method)
            condition1_intensity = np.concatenate([ rep['intensity'] for rep in data[condition_labels[0]].values() ])
            condition2_intensity = np.concatenate([ rep['intensity'] for rep in data[condition_labels[1]].values() ])
            condition1_dwell = np.concatenate([ rep['dwell'] for rep in data[condition_labels[0]].values() ])
            condition2_dwell = np.concatenate([ rep['dwell'] for rep in data[condition_labels[1]].values() ])

            for met in methods:
                logger.trace(f"Running {met} test on position {pos}")
                if met in ["MW", "KS", "TT"] :
                    try:
                        pvalues = nonparametric_test(condition1_intensity, condition2_intensity, condition1_dwell, condition2_dwell, method=met)
                    except:
                        raise NanocomporeError("Error doing {} test on reference {}".format(met, ref_id))
                    res["{}_intensity_pvalue".format(met)]=pvalues[0]
                    res["{}_dwell_pvalue".format(met)]=pvalues[1]
                    tests.add("{}_intensity_pvalue".format(met))
                    tests.add("{}_dwell_pvalue".format(met))
                elif met == "GMM":
                    try:
                        gmm_results = gmm_test(data, anova=anova, logit=logit, allow_warnings=allow_warnings, random_state=random_state)
                    except:
                        raise NanocomporeError("Error doing GMM test on reference {}".format(ref_id))
                    res["GMM_model"] = gmm_results['gmm']
                    if anova:
                        res["GMM_anova_pvalue"] = gmm_results['anova']['pvalue']
                        res["GMM_anova_model"] = gmm_results['anova']
                        tests.add("GMM_anova_pvalue")
                    if logit:
                        res["GMM_logit_pvalue"] = gmm_results['logit']['pvalue']
                        res["GMM_logit_model"] = gmm_results['logit']
                        tests.add("GMM_logit_pvalue")

            # Calculate shift statistics
            logger.trace(f"Calculatign shift stats for {pos}")
            res['shift_stats'] = shift_stats(condition1_intensity, condition2_intensity, condition1_dwell, condition2_dwell)
            # Save results in main
            logger.trace(f"Saving test results for {pos}")
            ref_pos_list[pos]['txComp'] = res
    logger.debug("Skipped {} positions because not present in all samples with sufficient coverage".format(n_lowcov))

    # Combine pvalue within a given sequence context
    if sequence_context > 0:
        logger.debug ("Calculate weighs and cross correlation matrices by tests")
        if sequence_context_weights == "harmonic":
            # Generate weights as a symmetrical harmonic series
            weights = harmomic_series(sequence_context)
        else:
            weights = [1]*(2*sequence_context+1)

        # Collect pvalue lists per tests
        pval_list_dict = defaultdict(list)
        for pos_dict in ref_pos_list:
            if 'txComp' in pos_dict:
                for test in tests:
                    pval_list_dict[test].append(pos_dict['txComp'][test])
            elif pos_dict["lowCov"]:
                for test in tests:
                    pval_list_dict[test].append(np.nan)
        # Compute cross correlation matrix per test
        corr_matrix_dict = OrderedDict()
        for test in tests:
            corr_matrix_dict[test] = cross_corr_matrix(pval_list_dict[test], sequence_context)

        logger.debug("Combine adjacent position pvalues with Hou's method position per position")
        # Iterate over each positions in previously generated result dictionary
        for mid_pos in range(len(ref_pos_list)):
            # Perform test only if middle pos is valid
            if not ref_pos_list[mid_pos]["lowCov"]:
                pval_list_dict = defaultdict(list)
                for pos in range(mid_pos-sequence_context, mid_pos+sequence_context+1):
                    for test in tests:
                        # If any of the positions is missing or any of the pvalues in the context is lowCov or NaN, consider it 1
                        if pos < 0 or pos >= len(ref_pos_list) or ref_pos_list[pos]["lowCov"] or np.isnan(ref_pos_list[pos]["txComp"][test]):
                            pval_list_dict[test].append(1)
                        # else just extract the corresponding pvalue
                        else:
                            pval_list_dict[test].append(ref_pos_list[pos]["txComp"][test])
                # Combine collected pvalues and add to dict
                for test in tests:
                    test_label = "{}_context_{}".format(test, sequence_context)
                    # If the mid p-value is.nan, force to nan also the context p-value
                    if np.isnan(ref_pos_list[mid_pos]["txComp"][test]):
                        ref_pos_list[mid_pos]['txComp'][test_label] = np.nan
                    else:
                        ref_pos_list[mid_pos]['txComp'][test_label] = combine_pvalues_hou(pval_list_dict[test], weights, corr_matrix_dict[test])

    return ref_pos_list

def nonparametric_test(condition1_intensity, condition2_intensity, condition1_dwell, condition2_dwell, method=None):

    if method in ["mann_whitney", "MW"]:
        stat_test = lambda x,y: mannwhitneyu(x, y, alternative='two-sided')
    elif method in ["kolmogorov_smirnov", "KS"]:
        stat_test = ks_twosamp
    elif method in ["t_test", "TT"]:
        stat_test = lambda x,y: ttest_ind(x, y, equal_var=False)
    else:
        raise NanocomporeError("Invalid statistical method name (MW, KS, ttest)")

    pval_intensity = stat_test(condition1_intensity, condition2_intensity)[1]
    if pval_intensity == 0: 
        pval_intensity = np.finfo(np.float).tiny

    pval_dwell = stat_test(condition1_dwell, condition2_dwell)[1]
    if pval_dwell == 0: 
        pval_dwell = np.finfo(np.float).tiny
    return(pval_intensity, pval_dwell)


def gmm_test(data, random_state, anova=True, logit=False, verbose=True, allow_warnings=False):
    # Condition labels
    condition_labels = tuple(data.keys())
    # List of sample labels
    sample_labels = list(data[condition_labels[0]].keys()) + list(data[condition_labels[1]].keys())

    if len(sample_labels) != len(set(sample_labels)):
        raise NanocomporeError("Sample labels have to be unique and it looks like some are not.")

    # Dictionary Sample_label:Condition_label
    sample_condition_labels = { sk:k for k,v in data.items() for sk in v.keys() }
    if len(condition_labels) != 2:
        raise NanocomporeError("gmm_test only supports two conditions")

    # Merge the intensities and dwell times of all samples in a single array
    global_intensity = np.concatenate(([v['intensity'] for v in data[condition_labels[0]].values()]+[v['intensity'] for v in data[condition_labels[1]].values()]), axis=None)
    global_dwell = np.concatenate(([v['dwell'] for v in data[condition_labels[0]].values()]+[v['dwell'] for v in data[condition_labels[1]].values()]), axis=None)
    global_dwell = np.log10(global_dwell)

    # Scale the intensity and dwell time arrays
    X = StandardScaler().fit_transform([(i, d) for i,d in zip(global_intensity, global_dwell)])

    # Generate an array of sample labels
    Y = [ k for k,v in data[condition_labels[0]].items() for _ in v['intensity'] ] + [ k for k,v in data[condition_labels[1]].items() for _ in v['intensity'] ]

    gmm_fit = fit_best_gmm(X, max_components=2, cv_types=['full'], random_state=random_state)
    gmm_mod, gmm_type, gmm_ncomponents = gmm_fit

    # If the best GMM has 2 clusters do an anova test on the log odd ratios
    if gmm_ncomponents == 2:
        # Assign data points to the clusters
        y_pred = gmm_mod.predict(X)
        counters = dict()
        # Count how many reads in each cluster for each sample
        for lab in sample_labels:
            counters[lab] = Counter(y_pred[[i==lab for i in Y]])
        cluster_counts = count_reads_in_cluster(counters)
        if anova:
            aov_results = gmm_anova_test(counters, sample_condition_labels, condition_labels, gmm_ncomponents, allow_warnings)
        else:
            aov_results=None

        if logit:
            logit_results = gmm_logit_test(Y, y_pred, sample_condition_labels, condition_labels)
        else:
            logit_results=None

    elif gmm_ncomponents == 1:
        aov_results = {'pvalue': np.nan, 'delta_logit': np.nan, 'table': "NC", 'cluster_counts': "NC"}
        logit_results = {'pvalue': np.nan, 'coef': "NC", 'model': "NC"}
        cluster_counts = "NC"
    else:
        raise NanocomporeError("GMM models with n_component>2 are not supported")

    return({'anova':aov_results, 'logit': logit_results, 'gmm':{'model': gmm_mod, 'cluster_counts': cluster_counts}})

def fit_best_gmm(X, random_state, max_components=2, cv_types=['spherical', 'tied', 'diag', 'full']):
   # Loop over multiple cv_types and n_components and for each fit a GMM
    # calculate the BIC and retain the lowest
    lowest_bic = np.infty
    bic = []
    n_components_range = range(1, max_components+1)
    for cv_type in cv_types:
        for n_components in n_components_range:
        # Fit a Gaussian mixture with EM
            gmm = GaussianMixture(n_components=n_components, covariance_type=cv_type, random_state=random_state)
            gmm.fit(X)
            bic.append(gmm.bic(X))
            if bic[-1] < lowest_bic:
                lowest_bic = bic[-1]
                best_gmm = gmm
                best_gmm_type = cv_type
                best_gmm_ncomponents = n_components
    return((best_gmm, best_gmm_type, best_gmm_ncomponents))

def gmm_anova_test(counters, sample_condition_labels, condition_labels, gmm_ncomponents, allow_warnings=False):
    labels= []
    logr = []
    for sample,counter in counters.items():
        # Save the condition label the corresponds to the current sample
        labels.append(sample_condition_labels[sample])
        # The Counter dictionaries in counters are not ordered
        # The following line enforces the order and adds 1 to avoid empty clusters
        ordered_counter = [ counter[i]+1 for i in range(gmm_ncomponents)]
        total = sum(ordered_counter)
        normalised_ordered_counter = [ i/total for i in ordered_counter ]
        # Loop through ordered_counter and divide each value by the first
        logr.append(np.log(normalised_ordered_counter[0]/(1-normalised_ordered_counter[0])))
    logr = np.around(np.array(logr), decimals=9)
    logr_s1 = [logr[i] for i,l in enumerate(labels) if l==condition_labels[0]]
    logr_s2 = [logr[i] for i,l in enumerate(labels) if l==condition_labels[1]]
    # If the SS for either array is 0, skip the anova test
    if sum_of_squares(logr_s1-np.mean(logr_s1)) == 0 and sum_of_squares(logr_s2-np.mean(logr_s2)) == 0:
        if not allow_warnings:
            raise NanocomporeError("While doing the Anova test we found a sample with within variance = 0. Use --allow_warnings to ignore.")
        else:
            aov_table = "Within variance is 0"
            aov_pvalue = np.finfo(np.float).tiny
    else:
        with warnings.catch_warnings():
            # Convert warnings to errors in order to catch them
            warnings.filterwarnings('error')
            try:
                aov_table = f_oneway(logr_s1, logr_s2)
                aov_pvalue = aov_table.pvalue
            except RuntimeWarning:
                if not allow_warnings:
                    raise NanocomporeError("While doing the Anova test a runtime warning was raised. Use --allow_warnings to ignore.")
                else:
                    warnings.filterwarnings('default')
                    aov_table = f_oneway(logr_s1, logr_s2)
                    aov_pvalue = np.finfo(np.float).tiny
    if aov_pvalue == 0:
        raise NanocomporeError("The Anova test returned a p-value of 0. This is most likely an error somewhere")
    # Calculate the delta log odds ratio, i.e. the difference of the means of the log odds ratios between the two conditions
    aov_delta_logit=float(np.mean(logr_s1)-np.mean(logr_s2))
    aov_results = {'pvalue': aov_pvalue, 'delta_logit': aov_delta_logit, 'table': aov_table, 'log_ratios':logr}
    return(aov_results)

def gmm_logit_test(Y, y_pred, sample_condition_labels, condition_labels):
    Y = [ sample_condition_labels[i] for i in Y]
    y_pred=np.append(y_pred, [0,0,1,1])
    Y.extend([condition_labels[0], condition_labels[1], condition_labels[0], condition_labels[1]])
    Y = pd.get_dummies(Y)
    Y['intercept']=1
    logit = dm.Logit(y_pred,Y[['intercept',condition_labels[1]]] )
    with warnings.catch_warnings():
        warnings.filterwarnings('error')
        try:
            logit_mod=logit.fit(disp=0)
            logit_pvalue, logit_coef = logit_mod.pvalues[1], logit_mod.params[1]
        except ConvergenceWarning:
            logit_mod, logit_pvalue, logit_coef = "NC", 1, "NC"
    if logit_pvalue == 0:
        logit_pvalue = np.finfo(np.float).tiny
    logit_results = {'pvalue': logit_pvalue, 'coef': logit_coef, 'model': logit_mod}
    return(logit_results)

def count_reads_in_cluster(counters):
    cluster_counts = list()
    for k,v in counters.items():
        cluster_counts.append("%s:%s/%s" % (k, v[0], v[1]))
    cluster_counts="__".join(cluster_counts)
    return(cluster_counts)

def shift_stats(condition1_intensity, condition2_intensity, condition1_dwell, condition2_dwell):
    """Calculate shift statistics"""
    shift_stats = OrderedDict([
        ('c1_mean_intensity', np.mean(condition1_intensity)),
        ('c2_mean_intensity', np.mean(condition2_intensity)),
        ('c1_median_intensity', np.median(condition1_intensity)),
        ('c2_median_intensity', np.median(condition2_intensity)),
        ('c1_sd_intensity', np.std(condition1_intensity)),
        ('c2_sd_intensity', np.std(condition2_intensity)),
        ('c1_mean_dwell', np.mean(condition1_dwell)),
        ('c2_mean_dwell', np.mean(condition2_dwell)),
        ('c1_median_dwell', np.median(condition1_dwell)),
        ('c2_median_dwell', np.median(condition2_dwell)),
        ('c1_sd_dwell', np.std(condition1_dwell)),
        ('c2_sd_dwell', np.std(condition2_dwell))
    ])
    return(shift_stats)


def cross_corr_matrix(pvalues_vector, context=2):
    """ Calculate the cross correlation matrix of the
        pvalues for a given context.
    """
    if len(pvalues_vector)<(context*3)+3:
        raise NanocomporeError("Not enough p-values for a context of order %s"%context)

    pvalues_vector = np.array([ i if not np.isnan(i) else 1 for i in pvalues_vector ])
    if any(pvalues_vector==0) or any(np.isinf(pvalues_vector)) or any(pvalues_vector>1):
        raise NanocomporeError("At least one p-value is invalid")

    matrix=[]
    s=pvalues_vector.size
    if all(p==1 for p in pvalues_vector):
        return(np.ones((context*2+1, context*2+1)))

    for i in range(-context,context+1):
        row=[]
        for j in range(-context,context+1):
            row.append(np.corrcoef((np.roll(pvalues_vector,i)[context:s-context]), (np.roll(pvalues_vector,j)[context:s-context]))[0][1])
        matrix.append(row)
    return(np.array(matrix))

def combine_pvalues_hou(pvalues, weights, cor_mat):
    """ Hou's method for the approximation for the distribution of the weighted
        combination of non-independent or independent probabilities.
        If any pvalue is nan, returns nan.
        https://doi.org/10.1016/j.spl.2004.11.028
        pvalues: list of pvalues to be combined
        weights: the weights of the pvalues
        cor_mat: a matrix containing the correlation coefficients between pvalues
        Test: when weights are equal and cor=0, hou is the same as Fisher
        print(combine_pvalues([0.1,0.02,0.1,0.02,0.3], method='fisher')[1])
        print(hou([0.1,0.02,0.1,0.02,0.3], [1,1,1,1,1], np.zeros((5,5))))
    """
    if(len(pvalues) != len(weights)):
        raise NanocomporeError("Can't combine pvalues is pvalues and weights are not the same length.")
    if( cor_mat.shape[0] != cor_mat.shape[1] or cor_mat.shape[0] != len(pvalues)):
        raise NanocomporeError("The correlation matrix needs to be squared, with each dimension equal to the length of the pvalued vector.")
    if all(p==1 for p in pvalues):
        return 1
    if any((p==0 or np.isinf(p) or p>1) for p in pvalues):
        raise NanocomporeError("At least one p-value is invalid")

    # Covariance estimation as in Kost and McDermott (eq:8)
    # https://doi.org/10.1016/S0167-7152(02)00310-3
    cov = lambda r: (3.263*r)+(0.710*r**2)+(0.027*r**3)
    k=len(pvalues)
    cov_sum=np.float64(0)
    sw_sum=np.float64(0)
    w_sum=np.float64(0)
    tau=np.float64(0)
    for i in range(k):
        for j in range(i+1,k):
            cov_sum += weights[i]*weights[j]*cov(cor_mat[i][j])
        sw_sum += weights[i]**2
        w_sum += weights[i]
        # Calculate the weighted Fisher's combination statistic
        tau += weights[i] * (-2*np.log(pvalues[i]))
    # Correction factor
    c = (2*sw_sum+cov_sum) / (2*w_sum)
    # Degrees of freedom
    f = (4*w_sum**2) / (2*sw_sum+cov_sum)
    # chi2.sf is the same as 1-chi2.cdf but is more accurate
    combined_p_value = chi2.sf(tau/c,f)
    # Return a very small number if pvalue = 0
    if combined_p_value == 0:
        combined_p_value = np.finfo(np.float).tiny
    return combined_p_value

def harmomic_series(sequence_context):
    weights = []
    for i in range(-sequence_context, sequence_context+1):
        weights.append(1/(abs(i)+1))
    return weights

def sum_of_squares(x):
    """
    Square each element of the input array and return the sum
    """
    x = np.atleast_1d(x)
    return np.sum(x*x)



# -*- coding: utf-8 -*-

#~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~#

# Standard library imports
import multiprocessing as mp
from time import time
from collections import *
import traceback
import os

# Third party imports
from loguru import logger
import numpy as np
from tqdm import tqdm

# Local imports
from nanocompore.common import *
from nanocompore.SuperParser import SuperParser

# Disable multithreading for MKL and openBlas
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["MKL_THREADING_LAYER"] = "sequential"
os.environ["NUMEXPR_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"
os.environ['OPENBLAS_NUM_THREADS'] = '1'

log_level_dict = {"debug":"DEBUG", "info":"INFO", "warning":"WARNING"}
logger.remove()

#~~~~~~~~~~~~~~MAIN CLASS~~~~~~~~~~~~~~#
class Eventalign_collapse ():

    def __init__ (self,
        eventalign_fn:str,
        outpath:str="./",
        outprefix:str="out",
        overwrite:bool = False,
        n_lines:int=None,
        nthreads:int = 3,
        progress:bool = False):
        """
        Collapse the nanopolish eventalign events at kmer level
        * eventalign_fn
            Path to a nanopolish eventalign tsv output file, or a list of file, or a regex (can be gzipped)
        * outpath
            Path to the output folder (will be created if it does exist yet)
        * outprefix
            text outprefix for all the files generated
        * overwrite
            If the output directory already exists, the standard behaviour is to raise an error to prevent overwriting existing data
            This option ignore the error and overwrite data if they have the same outpath and outprefix.
        * n_lines
            Maximum number of read to parse.
        * nthreads
            Number of threads (two are used for reading and writing, all the others for parallel processing).
        * progress
            Display a progress bar during execution
        """
        logger.info("Checking and initialising Eventalign_collapse")

        # Save init options in dict for later
        log_init_state(loc=locals())

        # Check threads number
        if nthreads < 3:
            raise NanocomporeError("Minimal required number of threads >=3")

        # Save args to self values
        self.__outpath = outpath
        self.__outprefix = outprefix
        self.__eventalign_fn = eventalign_fn
        self.__n_lines = n_lines
        self.__nthreads = nthreads - 2 # subtract 1 for reading and 1 for writing
        self.__progress = progress

        # Input file field selection typing and renaming
        self.__select_colnames = ["contig", "read_name", "position", "reference_kmer", "model_kmer", "event_length", "samples"]
        self.__change_colnames = {"contig":"ref_id" ,"position":"ref_pos", "read_name":"read_id", "samples":"sample_list", "event_length":"dwell_time"}
        self.__cast_colnames = {"ref_pos":int, "dwell_time":np.float32, "sample_list":lambda x: [float(i) for i in x.split(",")]}

    def __call__(self):
        """
        Run the analysis
        """
        logger.info("Starting data processing")
        # Init Multiprocessing variables
        in_q = mp.Queue(maxsize = 100)
        out_q = mp.Queue(maxsize = 100)
        error_q = mp.Queue()

        # Define processes
        ps_list = []
        ps_list.append (mp.Process (target=self.__split_reads, args=(in_q, error_q)))
        for i in range (self.__nthreads):
            ps_list.append (mp.Process (target=self.__process_read, args=(in_q, out_q, error_q)))
        ps_list.append (mp.Process (target=self.__write_output, args=(out_q, error_q)))

        # Start processes and monitor error queue
        try:
            # Start all processes
            for ps in ps_list:
                ps.start()

            # Monitor error queue
            for tb in iter(error_q.get, None):
                logger.trace("Error caught from error_q")
                raise NanocomporeError(tb)
            logger.debug("Caught poison pill in error queue")

            # Soft processes and queues stopping
            logger.debug("Joining all processes")
            for ps in ps_list:
                ps.join()
            logger.debug("All processes were joined")
            logger.debug("Closing all queues")
            for q in (in_q, out_q, error_q):
                q.close()
            logger.debug("All queues were closed")

        # Catch error, kill all processed and reraise error
        except Exception as E:
            logger.error("An error occured. Killing all processes and closing queues\n")
            try:
                for ps in ps_list:
                    ps.terminate()
                for q in (in_q, out_q, error_q):
                    q.close()
            except:
                logger.error("An error occured while trying to kill processes\n")
            raise E

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~PRIVATE METHODS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    def __split_reads (self, in_q, error_q):
        """
        Mono-threaded reader
        """
        logger.debug("Start reading input file(s)")

        n_reads = n_events = 0

        try:
            # Open input file with superParser
            with SuperParser(
                fn = self.__eventalign_fn,
                select_colnames = self.__select_colnames,
                cast_colnames = self.__cast_colnames,
                change_colnames = self.__change_colnames,
                n_lines=self.__n_lines) as sp:

                for l in sp:
                    n_events+=1

                    # First event exception
                    if n_events==1:
                        cur_ref_id = l["ref_id"]
                        cur_read_id = l["read_id"]
                        event_l = [l]

                    # Same read/ref group = just append to current event group
                    elif l["ref_id"] == cur_ref_id and l["read_id"] == cur_read_id:
                        event_l.append(l)

                    # If new read/ref group detected = enqueue previous event group and start new one
                    else:
                        n_reads+=1
                        in_q.put(event_l)

                        cur_ref_id = l["ref_id"]
                        cur_read_id = l["read_id"]
                        event_l = [l]

                # Last event line exception
                in_q.put(event_l)
                n_reads+=1

        # Manage exceptions and add error trackback to error queue
        except Exception:
            logger.debug("Error in Reader")
            error_q.put (NanocomporeError(traceback.format_exc()))

        # Deal poison pills
        finally:
            for i in range (self.__nthreads):
                in_q.put(None)
            logger.debug("Parsed Reads:{} Events:{}".format(n_reads, n_events))

    def __process_read (self, in_q, out_q, error_q):
        """
        Multi-threaded workers collapsing events at kmer level
        """
        logger.debug("Starting processing reads")
        try:
            n_reads = n_kmers = n_events = n_signals = 0

            # Take on event list corresponding to one read from the list
            for events_l in iter(in_q.get, None):

                # Create an empty Read object and fill it with event lines
                # events aggregation at kmer level is managed withon the object
                read = Read (read_id=events_l[0]["read_id"], ref_id=events_l[0]["ref_id"])
                for event_d in events_l:
                    read.add_event(event_d)
                    n_events+=1

                # If at least one valid event found collect results at read and kmer level
                if read.n_events > 1:
                    read_res_d = read.get_read_results()
                    kmer_res_l = read.get_kmer_results()
                    out_q.put((read_res_d, kmer_res_l))
                    n_reads+=1
                    n_kmers+= len(kmer_res_l)
                    n_signals+= read.n_signals

        # Manage exceptions and add error trackback to error queue
        except Exception:
            logger.error("Error in Worker. Event:{}".format(event_d))
            error_q.put (NanocomporeError(traceback.format_exc()))

        # Deal poison pill
        finally:
            logger.debug("Processed Reads:{} Kmers:{} Events:{} Signals:{}".format(n_reads, n_kmers, n_events, n_signals))
            out_q.put(None)

    def __write_output (self, out_q, error_q):
        """
        Mono-threaded Writer
        """
        logger.debug("Start writing output files")

        byte_offset = n_reads = n_kmers = 0

        # Init variables for index files
        idx_fn = os.path.join(self.__outpath, self.__outprefix+"_eventalign_collapse.tsv.idx")
        data_fn = os.path.join(self.__outpath, self.__outprefix+"_eventalign_collapse.tsv")

        try:
            # Open output files and tqdm progress bar
            with open (data_fn, "w") as data_fp, open (idx_fn, "w") as idx_fp, tqdm (unit=" reads", disable=not self.__progress) as pbar:

                # Iterate over out queue until nthread poison pills are found
                for _ in range (self.__nthreads):
                    for read_res_d, kmer_res_l in iter (out_q.get, None):
                        n_reads+=1

                        # Define file header from first read and first kmer
                        if byte_offset is 0:
                            idx_header_list = list(read_res_d.keys())+["byte_offset","byte_len"]
                            idx_header_str = "\t".join(idx_header_list)
                            data_header_list = list(kmer_res_l[0].keys())
                            data_header_str = "\t".join(data_header_list)

                            # Write index file header
                            idx_fp.write ("{}\n".format(idx_header_str))

                        # Write data file header
                        byte_len = 0
                        header_str = "#{}\t{}\n{}\n".format(read_res_d["read_id"], read_res_d["ref_id"], data_header_str)
                        data_fp.write(header_str)
                        byte_len+=len(header_str)

                        # Write kmer data matching data field order
                        for kmer in kmer_res_l:
                            n_kmers+=1
                            data_str = "\t".join([str(kmer[f]) for f in data_header_list])+"\n"
                            data_fp.write(data_str)
                            byte_len+=len(data_str)

                        # Add byte
                        read_res_d["byte_offset"] = byte_offset
                        read_res_d["byte_len"] = byte_len-1
                        idx_str = "\t".join([str(read_res_d[f]) for f in idx_header_list])
                        idx_fp.write("{}\n".format(idx_str))

                        # Update pbar
                        byte_offset+=byte_len
                        pbar.update(1)

                # Flag last line
                data_fp.write ("#\n")

        # Manage exceptions and add error trackback to error queue
        except Exception:
            logger.error("Error in Writer")
            error_q.put (NanocomporeError(traceback.format_exc()))

        finally:
            logger.debug("Written Reads:{} Kmers:{}".format(n_reads, n_kmers))
            logger.info ("Output reads written:{}".format(n_reads))
            # Kill error queue with poison pill
            error_q.put(None)

#~~~~~~~~~~~~~~~~~~~~~~~~~~HELPER CLASSES~~~~~~~~~~~~~~~~~~~~~~~~~~#

class Read ():
    """Helper class representing a single read"""

    def __init__ (self, read_id, ref_id):
        self.read_id = read_id
        self.ref_id = ref_id
        self.ref_start = None
        self.ref_end = None
        self.kmer_l = [Kmer()]
        self.n_events = 0
        self.n_signals = 0
        self.dwell_time = 0

    def __repr__(self):
        s = "Read instance\n"
        s+="\tread_id: {}\n".format(self.read_id)
        s+="\tref_id: {}\n".format(self.ref_id)
        s+="\tref_start: {}\n".format(self.ref_start)
        s+="\tref_end: {}\n".format(self.ref_end)
        s+="\tn_events: {}\n".format(self.n_events)
        s+="\tdwell_time: {}\n".format(self.dwell_time)
        for status, count in self.kmers_status.items():
            s+="\t{}: {}\n".format(status, count)
        return s

    def add_event (self, event_d):
        self.n_events+=1
        self.n_signals+=len(event_d["sample_list"])
        self.dwell_time+=event_d["dwell_time"]

        # First event
        if not self.ref_end:
            self.ref_start = event_d["ref_pos"]
            self.ref_end = event_d["ref_pos"]

        # Position offset = Move to next position
        if event_d["ref_pos"] > self.ref_end:
            self.kmer_l.append(Kmer())
            self.ref_end = event_d["ref_pos"]
            self.ref_end = event_d["ref_pos"]

        # Update current kmer
        self.kmer_l[-1].add_event(event_d)

    @property
    def kmers_status (self):
        d = OrderedDict()
        d["kmers"] = self.ref_end-self.ref_start+1
        d["missing_kmers"] = d["kmers"] - len(self.kmer_l)
        d["NNNNN_kmers"]=0
        d["mismatch_kmers"]=0
        d["valid_kmers"]=0
        for k in self.kmer_l:
            d[k.status+"_kmers"]+=1
        return d

    def get_read_results (self):
        d = OrderedDict()
        d["ref_id"] = self.ref_id
        d["ref_start"] = self.ref_start
        d["ref_end"] = self.ref_end
        d["read_id"] = self.read_id
        d["num_events"] = self.n_events
        d["num_signals"] = self.n_signals
        d["dwell_time"] = self.dwell_time
        for status, count in self.kmers_status.items():
            d[status] = count
        return d

    def get_kmer_results (self):
        l = [kmer.get_results() for kmer in self.kmer_l]
        return l

class Kmer ():
    """Helper class representing a single kmer"""

    def __init__ (self):
        self.ref_pos = None
        self.ref_kmer = None
        self.num_events = 0
        self.num_signals = 0
        self.dwell_time = 0.0
        self.NNNNN_dwell_time = 0.0
        self.mismatch_dwell_time = 0.0
        self.sample_list = []

    def __repr__(self):
        s = "Kmer instance\n"
        s+="\tref_pos: {}\n".format(self.ref_pos)
        s+="\tref_kmer: {}\n".format(self.ref_kmer)
        s+="\tnum_events: {}\n".format(self.num_events)
        s+="\tdwell_time: {}\n".format(self.dwell_time)
        s+="\tNNNNN_dwell_time: {}\n".format(self.NNNNN_dwell_time)
        s+="\tmismatch_dwell_time: {}\n".format(self.mismatch_dwell_time)
        s+="\tsample_list: {}\n".format(" ".join([str(i) for i in self.sample_list]))
        return s

    def add_event (self, event_d):
        # First event
        if not self.ref_pos:
            self.ref_pos = event_d["ref_pos"]
            self.ref_kmer = event_d["reference_kmer"]

        # Update kmer values
        self.num_events+=1
        self.num_signals+=len(event_d["sample_list"])
        self.dwell_time+=event_d["dwell_time"]
        if event_d["model_kmer"] == "NNNNN":
            self.NNNNN_dwell_time+=event_d["dwell_time"]
        elif event_d["model_kmer"] != self.ref_kmer:
            self.mismatch_dwell_time+=event_d["dwell_time"]
        self.sample_list.extend(event_d["sample_list"])

    @property
    def status (self):
        """Define kmer status depending on mismatching or NNNNN events"""
        # If no NNNNN nor mismatch events then the kmer the purely valid
        if not self.NNNNN_dwell_time and not self.mismatch_dwell_time:
            return "valid"
        # Otherwise return status corresponding to longuest invalid dwell time
        elif self.NNNNN_dwell_time > self.mismatch_dwell_time:
            return "NNNNN"
        else:
            return "mismatch"

    def get_results(self):
        sample_array = np.array(self.sample_list, dtype=np.float32)
        d = OrderedDict()
        d["ref_pos"] = self.ref_pos
        d["ref_kmer"] = self.ref_kmer
        d["num_events"] = self.num_events
        d["num_signals"] = self.num_signals
        d["dwell_time"] = self.dwell_time
        d["NNNNN_dwell_time"] = self.NNNNN_dwell_time
        d["mismatch_dwell_time"] = self.mismatch_dwell_time
        d["status"] = self.status
        d["median"] = np.median(sample_array)
        d["mad"] = np.median(np.abs(sample_array-d["median"]))

        return d



# -*- coding: utf-8 -*-

#~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~#
# Standard library imports
import os
from collections import *
import gzip
from glob import iglob
from loguru import logger

#~~~~~~~~~~~~~~CLASS~~~~~~~~~~~~~~#

class SuperParser ():
    def __init__ (self,
        fn,
        select_colnames=[],
        n_lines=None,
        sep="\t",
        comment="#",
        change_colnames={},
        cast_colnames={}):
        """
        Open a parser for field delimited file and return an iterator yield lines as namedtuples
        Transparently parse gziped and multiple file with the same header
        * fn
            Path to a field delimited file or a list of files or a regex or a mix of everything (Can be gzipped)
        * select_colnames
            List of column names to use parse and return
        * sep
            field separator
        * comment
            skip any line starting with this string
        * cast_colnames
            Dict corresponding to fields (based on colnames) to cast in a given python type
        """

        # Init logger and counter
        self.counter = Counter()

        # Save self variables
        self._sep = sep
        self._comment = comment
        self._n_lines = n_lines

        # Input file opening
        self.f_list = self._open_files (fn)

        # Define colnames based on file header. It needs to be the same for all the files to parse
        fn, fp = self.f_list[0]
        logger.debug("Reading header from file: {}".format(fn))
        self.colnames = self._get_first_line_header(fp)
        self.select_idx = []

        # Find selected colnames if needed
        if select_colnames:
            for field in select_colnames:
                try:
                    idx = self.colnames.index(field)
                    self.select_idx.append(idx)
                except ValueError:
                    raise SuperParserError ("Required field `{}` not found in files `{}`".format(field, fn))
            self.colnames = select_colnames

        # Check header for subsequent files if any
        if len(self.f_list)>1:
            for fn, fp in self.f_list[1:]:
                colnames = self._get_first_line_header(fp)
                if self.select_idx:
                    for idx, col in zip(self.select_idx, self.colnames):
                        if idx >= len(colnames) or colnames[idx] != col:
                            raise SuperParserError ("Inconsistant headers between `{}` and `{}`".format(self.f_list[0][0], fn))
                else:
                    if colnames != self.colnames:
                        raise SuperParserError ("Inconsistant headers between `{}` and `{}`".format(self.f_list[0][0], fn))

        # Change colnames is required
        if change_colnames:
            for i, col in enumerate(self.colnames):
                if col in change_colnames:
                    self.colnames[i] = change_colnames[col]

        logger.debug("Column names from header: `{}`".format(" / ".join(self.colnames)))

        # Save initial number of columns, custom namedtuple and dtype index
        self.ncols = len(self.colnames)
        self.lt = namedtuple("lt", self.colnames)
        self.cast_colnames = cast_colnames

    #~~~~~~~~~~~~~~MAGIC AND PROPERTY METHODS~~~~~~~~~~~~~~#

    def __len__ (self):
        size = 0
        for fn, fp in self.f_list:
            size+= int(os.path.getsize(fn))
        return size-self._header_len

    def __enter__ (self):
        return self

    def close (self):
        for i, j in self.counter.most_common():
            logger.debug("{}: {}".format(i, j))
        for fn, fp in self.f_list:
            try:
                logger.debug ("Closing file:{}".format(fn))
                fp.close()
            except Exception as E:
                logger.warning (E)

    def __exit__(self, exception_type, exception_val, trace):
        self.close()

    def __iter__ (self):
        # Iterate over files
        for fn, fp in self.f_list:
            logger.debug("Starting to parse file {}".format(fn))
            # Iterate over line in file
            for line in fp:
                self.counter["Lines Parsed"]+=1

                if self._comment and line.startswith(self._comment):
                    self.counter["Comment lines skipped"]+=1
                    continue
                try:
                    line = self._parse_line(line)
                    self.counter["Lines successfully parsed"]+=1
                    yield line
                    # early stopping condition
                    if self._n_lines and self.counter["Lines successfully parsed"] == self._n_lines:
                        return
                except (SuperParserError, TypeError) as E:
                    self.counter["Malformed or Invalid Lines"]+=1
                    logger.debug(E)
                    logger.debug("File {}: Invalid line {}".format(fn, line))
            logger.debug("End of file: {}".format(fn))
        logger.debug("All files done")

    #~~~~~~~~~~~~~~PRIVATE METHODS~~~~~~~~~~~~~~#

    def _get_first_line_header (self, fp):
        header_line = next(fp)
        header_list = header_line.rstrip().split(self._sep)
        return header_list

    def _parse_line (self, line):

        # Split line
        line = line.rstrip().split(self._sep)

        # Select field if needed
        if self.select_idx:
            line = [line[i] for i in self.select_idx]

        # Raise error if the length of the line is inconsistent with the header
        if len(line) != self.ncols:
            raise SuperParserError("Invalid Number of fields found")
        line_d = OrderedDict(zip(self.colnames,line))

        # Cast values according to provided types of lambda functions
        if self.cast_colnames:
            for colname, cast in self.cast_colnames.items():
                try:
                    line_d[colname] = cast(line_d[colname])
                except Exception:
                    raise SuperParserError("Cannot cast field in required type")

        # Return parsed line as a dict
        return line_d

    def _open_files (self, fn_list):
        """Transparently open files, lists, regex, gzipped or not"""
        f_list = []

        # Standard input
        if fn_list is 0:
            fn = "stdin"
            fp = open(0)
            return [(fn,fp)]

        # Cast single file or regex to list
        if isinstance(fn_list, str):
            fn_list = [fn_list]

        if isinstance(fn_list, (list, tuple, set)):
            for fn_regex in fn_list:
                for fn in iglob(fn_regex):
                    self.counter["Input files"]+=1
                    if fn.endswith(".gz"):
                        logger.debug("Opening file {} in gzip mode".format(fn))
                        fp = gzip.open(fn, "rt")
                    else:
                        logger.debug("Opening file {} in normal mode".format(fn))
                        fp = open(fn, "r")
                    f_list.append((fn,fp))

            return f_list

        else:
            raise SuperParserError ("Invalid file type")

class SuperParserError (Exception):
    """ Basic exception class for SuperParserError """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~#

# Standard library imports
import argparse
from collections import *
import textwrap
import os

# Third party
from loguru import logger

# Local imports
from nanocompore import __version__ as package_version
from nanocompore import __name__ as package_name
from nanocompore import __description__ as package_description
from nanocompore.SampComp import SampComp
from nanocompore.SimReads import SimReads
from nanocompore.Eventalign_collapse import Eventalign_collapse
from nanocompore.common import *

#~~~~~~~~~~~~~~MAIN PARSER ENTRY POINT~~~~~~~~~~~~~~#

def main(args=None):
    # General parser
    parser = argparse.ArgumentParser(description=package_description, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--version', '-v', action='version', version='v'+package_version)
    subparsers = parser.add_subparsers(dest='subcommand',
        description=textwrap.dedent("""
            nanocompore implements the following subcommands\n
            \t* eventalign_collapse : Collapse the nanopolish eventalign output at kmers level and compute kmer level statistics\n
            \t* sampcomp : Compare 2 samples and find significant signal differences\n
            \t* simreads : Simulate reads as a NanopolishComp like file from a fasta file and an inbuild model"""))
    subparsers.required = True

    # Sampcomp subparser
    parser_sc = subparsers.add_parser('sampcomp', formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent("""
            Compare 2 samples and find significant signal differences\n
            * Minimal example with file_list arguments\n
                nanocompore sampcomp -1 f1.tsv,f2.tsv -2 f3.tsv,f4.tsv -f ref.fa -o results
            * Minimal example with sample YAML file\n
                nanocompore sampcomp -y samples.yaml -f ref -o results"""))
    parser_sc.set_defaults(func=sampcomp_main)
    parser_sc_sample_yaml = parser_sc.add_argument_group('YAML sample files', description="Option allowing to describe sample files in a YAML file")
    parser_sc_sample_yaml.add_argument("--sample_yaml", "-y", default=None, type=str, metavar="sample_yaml",
        help="YAML file containing the sample file labels. See formatting in documentation. (required if --file_list1 and --file_list2 not given)")
    parser_sc_sample_args = parser_sc.add_argument_group('Arguments sample files', description="Option allowing to describe sample files directly as command line arguments")
    parser_sc_sample_args.add_argument("--file_list1", "-1", default=None, type=str, metavar="/path/to/Condition1_rep1,/path/to/Condition1_rep2",
        help="Comma separated list of NanopolishComp files for label 1. (required if --sample_yaml not given)")
    parser_sc_sample_args.add_argument("--file_list2", "-2", default=None, type=str, metavar="/path/to/Condition2_rep1,/path/to/Condition2_rep2",
        help="Comma separated list of NanopolishComp files for label 2. (required if --sample_yaml not given)")
    parser_sc_sample_args.add_argument("--label1", type=str, metavar="Condition1", default="Condition1",
        help="Label for files in --file_list1 (default: %(default)s)")
    parser_sc_sample_args.add_argument("--label2", type=str, metavar="Condition2", default="Condition2",
        help="Label for files in --file_list2 (default: %(default)s)")
    parser_sc_io = parser_sc.add_argument_group('Input options')
    parser_sc_io.add_argument("--fasta", "-f", type=str, required=True,
        help="Fasta file used for mapping (required)")
    parser_sc_io.add_argument("--bed", type=str, default=None,
        help="BED file with annotation of transcriptome used for mapping (optional)")
    parser_sc_filtering = parser_sc.add_argument_group('Transcript filtering options')
    parser_sc_filtering.add_argument("--max_invalid_kmers_freq", type=float, default=0.1,
        help="Max fequency of invalid kmers (default: %(default)s)")
    parser_sc_filtering.add_argument("--min_coverage", type=int, default=30,
        help="Minimum coverage required in each condition to do the comparison (default: %(default)s)")
    parser_sc_filtering.add_argument("--downsample_high_coverage", type=int, default=5000,
        help="Transcripts with high coverage will be downsampled (default: %(default)s)")
    parser_sc_filtering.add_argument("--min_ref_length", type=int, default=100,
        help="Minimum length of a reference transcript to include it in the analysis (default: %(default)s)")
    parser_sc_testing = parser_sc.add_argument_group('Statistical testing options')
    parser_sc_testing.add_argument("--comparison_methods", type=str, default="GMM,KS",
        help="Comma separated list of comparison methods. Valid methods are: GMM,KS,TT,MW. (default: %(default)s)")
    parser_sc_testing.add_argument("--sequence_context", type=int, default=0, choices=range(0,5),
        help="Sequence context for combining p-values (default: %(default)s)")
    parser_sc_testing.add_argument("--sequence_context_weights", type=str, default="uniform", choices=["uniform", "harmonic"],
        help="Type of weights to use for combining p-values")
    parser_sc_testing.add_argument("--pvalue_thr", type=float, default=0.05,
        help="Adjusted p-value threshold for reporting significant sites (default: %(default)s)")
    parser_sc_testing.add_argument("--logit", action='store_true',
        help="Use logistic regression testing downstream of GMM method. This is a legacy option and is now the deault.")
    parser_sc_testing.add_argument("--anova", action='store_true',
        help="Use Anova test downstream of GMM method (default: %(default)s)")
    parser_sc_testing.add_argument("--allow_warnings", action='store_true', default=False,
        help="If True runtime warnings during the ANOVA tests don't raise an error (default: %(default)s)")
    parser_sc_misc = parser_sc.add_argument_group('Other options')
    parser_sc_misc.add_argument("--nthreads", "-t", type=int, default=3,
        help="Number of threads (default: %(default)s)")

    # simreads subparser
    parser_sr = subparsers.add_parser('simreads', formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent("""
            Simulate reads as a NanopolishComp like file from a fasta file and an inbuild model\n
            * Minimal example without model alteration
                nanocompore simreads -f ref.fa -o results -n 50\n
            * Minimal example with alteration of model intensity loc parameter for 50% of the reads
            nanocompore simreads -f ref.fa -o results -n 50 --intensity_mod 2 --mod_reads_freq 0.5 --mod_bases_freq 0.2"""))
    parser_sr.set_defaults(func=simreads_main)
    parser_sr_io = parser_sr.add_argument_group('Input options')
    parser_sr_io.add_argument("--fasta", "-f", type=str, required=True,
        help="Fasta file containing references to use to generate artificial reads")
    parser_sr_modify = parser_sr.add_argument_group('Signal modification options')
    parser_sr_modify.add_argument("--intensity_mod", type=float, default=0,
        help="Fraction of intensity distribution SD by which to modify the intensity distribution loc value (default: %(default)s)")
    parser_sr_modify.add_argument("--dwell_mod", type=float, default=0,
        help="Fraction of dwell time distribution SD by which to modify the intensity distribution loc value (default: %(default)s)")
    parser_sr_modify.add_argument("--mod_reads_freq", type=float, default=0,
        help="Frequency of reads to modify (default: %(default)s)")
    parser_sr_modify.add_argument("--mod_bases_freq", type=float, default=0.25,
        help="Frequency of bases to modify in each read (if possible) (default: %(default)s)")
    parser_sr_modify.add_argument("--mod_bases_type", type=str, default="A", choices=["A","T","C","G"],
        help="Base for which to modify the signal (default: %(default)s)")
    parser_sr_modify.add_argument("--mod_extend_context", type=int, default=2,
        help="number of adjacent base affected by the signal modification following an harmonic series (default: %(default)s)")
    parser_sr_modify.add_argument("--min_mod_dist", type=int, default=6,
        help="Minimal distance between 2 bases to modify (default: %(default)s)")
    parser_sr_misc = parser_sr.add_argument_group('Other options')
    parser_sr_misc.add_argument("--run_type", type=str, default="RNA", choices=["RNA", "DNA"],
        help="Define the run type model to import (default: %(default)s)")
    parser_sr_misc.add_argument("--nreads_per_ref", "-n", type=int, default=100,
        help="Number of reads to generate per references (default: %(default)s)")
    parser_sr_misc.add_argument("--pos_rand_seed", type=int, default=42 ,
        help="Define a seed for randon position picking to get a deterministic behaviour (default: %(default)s)")
    parser_sr_misc.add_argument("--not_bound", action='store_true', default=False,
        help="Do not bind the values generated by the distributions to the observed min and max observed values from the model file (default: %(default)s)")

    # Eventalign_collapse subparser
    parser_ec = subparsers.add_parser("eventalign_collapse", formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent("""
        Collapse the nanopolish eventalign output at kmers level and compute kmer level statistics
        * Minimal example
            nanocompore eventalign_collapse -i nanopolish_eventalign.tsv -outprefix out\n"""))
    parser_ec.set_defaults(func=eventalign_collapse_main)
    parser_ec_io = parser_ec.add_argument_group("Input options")
    parser_ec_io.add_argument("--eventalign", "-i", default=0,
        help="Path to a nanopolish eventalign tsv output file, or a list of file, or a regex (can be gzipped). It can be ommited if piped to standard input (default: piped to stdin)")
    parser_ec_rp = parser_ec.add_argument_group("Run parameters options")
    parser_ec_rp.add_argument("--n_lines", "-l", default=None , type=int ,
        help = "Number of lines to parse.(default: no limits")
    parser_ec_misc = parser_ec.add_argument_group("Other options")
    parser_ec_misc.add_argument("--nthreads", "-t", default=3, type=int,
        help="Total number of threads. 2 threads are reserved for the reader and the writer (default: %(default)s)")

    # Add common options for all parsers
    for sp in [parser_sc, parser_sr, parser_ec]:
        sp_output = sp.add_argument_group("Output options")
        sp_output.add_argument("--outpath", "-o", type=str, default="./",
            help="Path to the output folder (default: %(default)s)")
        sp_output.add_argument("--outprefix", "-p", type=str, default="out",
            help="text outprefix for all the files generated (default: %(default)s)")
        sp_output.add_argument("--overwrite", "-w", action='store_true', default=False,
            help="Use --outpath even if it exists already (default: %(default)s)")
        sp_verbosity = sp.add_argument_group("Verbosity options")
        sp_verbosity.add_argument("--log_level", type=str, default="info", choices=["warning", "info", "debug", "trace"],
            help="Set the log level (default: %(default)s)")
        sp_verbosity.add_argument("--progress", default=False, action='store_true',
            help="Display a progress bar during execution (default: %(default)s)")

    # Parse agrs and
    args = parser.parse_args()

    # Check if output folder already exists
    try:
        mkdir(fn=args.outpath, exist_ok=args.overwrite)
    except (NanocomporeError, FileExistsError) as E:
        raise NanocomporeError("Could not create the output folder. Try using `--overwrite` option or use another directory")

    # Set logger
    log_fn = os.path.join(args.outpath, args.outprefix+"_{}.log".format(vars(args)["subcommand"]))
    set_logger(args.log_level, log_fn=log_fn)

    # Call relevant subfunction
    args.func(args)

#~~~~~~~~~~~~~~SUBCOMMAND FUNCTIONS~~~~~~~~~~~~~~#

def sampcomp_main(args):
    """"""
    logger.warning("Running SampComp")

    # Load eventalign_fn_dict from a YAML file or assemble eventalign_fn_dict for the command line option
    if args.sample_yaml:
        eventalign_fn_dict = args.sample_yaml
    elif args.file_list1 and args.file_list2:
        eventalign_fn_dict = build_eventalign_fn_dict(args.file_list1, args.file_list2, args.label1, args.label2)
    else:
        raise NanocomporeError("Samples eventalign files have to be provided with either `--sample_yaml` or `--file_list1` and `--file_list2`")

    # Init SampComp
    s = SampComp(
        eventalign_fn_dict = eventalign_fn_dict,
        max_invalid_kmers_freq = args.max_invalid_kmers_freq,
        outpath = args.outpath,
        outprefix = args.outprefix,
        overwrite = args.overwrite,
        fasta_fn = args.fasta,
        bed_fn = args.bed,
        nthreads = args.nthreads,
        min_coverage = args.min_coverage,
        min_ref_length = args.min_ref_length,
        downsample_high_coverage = args.downsample_high_coverage,
        comparison_methods = args.comparison_methods,
        logit = True,
        anova = args.anova,
        allow_warnings = args.allow_warnings,
        sequence_context = args.sequence_context,
        sequence_context_weights = args.sequence_context_weights,
        progress = args.progress)

    # Run SampComp
    db = s()

    # Save all reports
    logger.info("Saving results")
    if(db):
        db.save_all(pvalue_thr=args.pvalue_thr)

def simreads_main(args):
    """"""
    logger.warning("Running SimReads")

    # Run SimReads
    SimReads(
        fasta_fn = args.fasta,
        outpath = args.outpath,
        outprefix = args.outprefix,
        overwrite = args.overwrite,
        run_type = args.run_type,
        nreads_per_ref = args.nreads_per_ref,
        intensity_mod = args.intensity_mod,
        dwell_mod = args.dwell_mod,
        mod_reads_freq = args.mod_reads_freq,
        mod_bases_freq = args.mod_bases_freq,
        mod_bases_type = args.mod_bases_type,
        mod_extend_context = args.mod_extend_context,
        min_mod_dist = args.min_mod_dist,
        pos_rand_seed = args.pos_rand_seed,
        not_bound = args.not_bound,
        progress = args.progress)

def eventalign_collapse_main (args):
    """"""
    logger.warning("Running Eventalign_collapse")

    # Init Eventalign_collapse
    e = Eventalign_collapse (
        eventalign_fn = args.eventalign,
        outpath = args.outpath,
        outprefix = args.outprefix,
        overwrite = args.overwrite,
        n_lines = args.n_lines,
        nthreads = args.nthreads,
        progress = args.progress)

    # Run eventalign_collapse
    e()

#~~~~~~~~~~~~~~CLI ENTRYPOINT~~~~~~~~~~~~~~#

if __name__ == "__main__":
    # execute only if run as a script
    main()



# -*- coding: utf-8 -*-

#~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~#
# Std lib
from collections import *
import logging
from pkg_resources import resource_filename
import json
import datetime
import os

# Third party
import numpy as np
import pandas as pd
from pyfaidx import Fasta
import pyfaidx
from scipy.stats import logistic as sp_logistic
from scipy.stats import wald as sp_wald
import matplotlib.pyplot as pl
from tqdm import tqdm

# Local package
from nanocompore.common import *

#~~~~~~~~~~~~~~MAIN CLASS~~~~~~~~~~~~~~#
def SimReads(
    fasta_fn:str,
    outpath:str = "./",
    outprefix:str = "out",
    overwrite:bool = False,
    run_type:str= "RNA",
    ref_list:list = [],
    nreads_per_ref:int=100,
    plot:bool=False,
    intensity_mod:float = 0,
    dwell_mod:float = 0,
    mod_reads_freq:float = 0,
    mod_bases_freq:float = 0.25,
    mod_bases_type:str = "A",
    mod_extend_context:int = 2,
    min_mod_dist:int = 6,
    pos_rand_seed:int = 42,
    data_rand_seed:int = None,
    not_bound:bool = False,
    progress:bool = False):
    """
    Simulate reads in a NanopolishComp like file from a fasta file and an inbuild model.
    The simulated reads correspond to the sequences provided in the fasta file and follow
    the intensity and dwell time from the corresponding model (RNA or DNA).
    * fasta_fn
        Fasta file containing references to use to generate artificial reads.
    * outpath
        Path to the output folder.
    * outprefix
        text outprefix for all the files generated by the function.
    * overwrite
        If the output directory already exists, the standard behaviour is to raise an error to prevent overwriting existing data
        This option ignore the error and overwrite data if they have the same outpath and outprefix.
    * run_type
        Define the run type model to import {RNA,DNA}
    * ref_list
        Restrict the references to the listed IDs.
    * nreads_per_ref
        Number of reads to generate per references.
    * plot
        If true, generate an interactive plot of the trace generated.
    * intensity_mod
        Fraction of intensity distribution SD by which to modify the intensity distribution loc value.
    * dwell_mod
        Fraction of dwell time distribution SD by which to modify the intensity distribution loc value.
    * mod_reads_freq
        Frequency of reads to modify.
    * mod_bases_freq
        Frequency of bases to modify in each read (if possible).
    * mod_bases_type
        Base for which to modify the signal. {A,T,C,G}
    * mod_extend_context
        number of adjacent base affected by the signal modification following an harmonic series.
    * min_mod_dist
        Minimal distance between 2 bases to modify.
    * pos_rand_seed
        Define a seed for randon position picking to get a deterministic behaviour.
    * data_rand_seed
        Define a seed for generating the data. If None (default) the seed is drawn from /dev/urandom.
    * not_bound
        Do not bind the values generated by the distributions to the observed min and max observed values from the model file.
    * progress
        Display a progress bar during execution
    """
    logger.info("Checking and initialising Simreads")

    # Save init options in dict for later
    log_init_state(loc=locals())

    # Check if fasta file exists
    if not access_file(fasta_fn):
        raise NanocomporeError("{} is not a valid file".format(fasta_fn))

    # Define model depending on run_type
    if run_type == "RNA":
        logger.info("Importing RNA model file")
        model_fn = resource_filename("nanocompore", "models/kmers_model_RNA_r9.4_180mv.tsv")
        model_df = pd.read_csv(model_fn, sep="\t", comment="#", index_col=0)
    else:
        raise NanocomporeError("Only RNA is implemented at the moment")

    # Open fasta file and output files
    logger.info("Reading Fasta file and simulate corresponding data")

    with pyfaidx.Fasta(fasta_fn) as fasta_fp,\
         open(os.path.join(outpath, "{}.tsv".format(outprefix)) , "w") as data_fp,\
         open(os.path.join(outpath, "{}.tsv.idx".format(outprefix)), "w") as idx_fp,\
         open(os.path.join(outpath, "{}_pos.tsv".format(outprefix)), "w") as pos_fp:

        # Get all reference names if no ref_list
        if not ref_list:
            ref_list = fasta_fp.keys()

        # Write log and index file header and init
        idx_fp.write("ref_id\tread_id\tbyte_offset\tbyte_len\n")
        pos_fp.write("ref_id\tmodified_positions\n")

        byte_offset = 0
        # Simulate reference per reference
        for ref_num, ref_id in enumerate(tqdm((ref_list), unit=" References", disable=not progress)):
            logger.debug("Processing reference {}".format(ref_id))
            try:
                ref_seq = str(fasta_fp[ref_id])

                # Simulate data corresponding to the reference
                intensity_array, dwell_array, mod_pos_list, nreads_mod = simulate_ref_mod_context(
                    ref_seq = ref_seq,
                    model_df = model_df,
                    nreads = nreads_per_ref,
                    intensity_mod = intensity_mod,
                    dwell_mod = dwell_mod,
                    mod_reads_freq = mod_reads_freq,
                    mod_bases_freq = mod_bases_freq,
                    mod_bases_type = mod_bases_type,
                    mod_extend_context = mod_extend_context,
                    min_mod_dist = min_mod_dist,
                    pos_rand_seed=pos_rand_seed,
                    data_rand_seed=data_rand_seed,
                    not_bound=not_bound)

                # Plot traces if required
                if plot:
                    plot_trace(ref_id, intensity_array, dwell_array, mod_pos_list, nreads_mod)

                # Write options used in log file
                pos_fp.write("{}\t{}\n".format(ref_id, array_join(";", mod_pos_list)))

                # Write output in NanopolishComp like files
                for read_num in range(nreads_per_ref):
                    read_str = "#{}_{}\t{}\n".format(ref_num, read_num, ref_id)
                    read_str += "ref_pos\tref_kmer\tdwell_time\tmedian\n"
                    for ref_pos in range(len(ref_seq)-4):
                        read_str += "{}\t{}\t{}\t{}\n".format(
                            ref_pos,
                            ref_seq[ref_pos:ref_pos+5],
                            dwell_array [ref_pos, read_num],
                            intensity_array [ref_pos, read_num])

                    data_fp.write(read_str)
                    idx_fp.write("{}\t{}_{}\t{}\t{}\n".format(ref_id, ref_num, read_num, byte_offset, len(read_str)-1))
                    byte_offset += len(read_str)

            except KeyError:
                logger.debug("Reference {} not found in reference fasta file".format(ref_id))

def plot_trace(ref_id, intensity_array, dwell_array, mod_pos_list, nreads_mod):
    """"""
    with pl.style.context("ggplot"):
        fig, axes = pl.subplots(2, 1, figsize=(30,10))

        # Plot intensity data
        for i, line in enumerate(intensity_array.T):
            axes[0].plot(line, alpha=(1/len(intensity_array.T))*2, color="red" if i<nreads_mod else "black")
        axes[0].set_title("Median intensity")
        axes[0].set_xlim(0,len(intensity_array))

        # Plot dwell data
        for i, line in enumerate(dwell_array.T):
            axes[1].plot(line, alpha=(1/len(dwell_array.T))*2, color="red" if i<nreads_mod else "black")
        axes[1].set_title("Dwell time")
        axes[1].set_xlim(0,len(dwell_array))

        # Add lines where the signal is modified
        for x in mod_pos_list:
            axes[0].axvline(x, color="grey", linestyle=":")
            axes[1].axvline(x, color="grey", linestyle=":")

        # tweak plot
        fig.suptitle(ref_id, y=1.02, fontsize=18)
        fig.tight_layout()

def simulate_ref_mod_context(
    ref_seq,
    model_df,
    nreads=100,
    intensity_mod=0,
    dwell_mod=0,
    mod_reads_freq=0,
    mod_bases_freq=0,
    mod_bases_type="A",
    mod_extend_context=0,
    min_mod_dist=6,
    not_bound=False,
    pos_rand_seed=42,
    data_rand_seed=None):
    """"""

    # Create empty arrays to store reads intensity and dwell
    n_kmers = len(ref_seq)-4
    intensity_array = np.empty(shape=(n_kmers, nreads), dtype=np.float)
    dwell_array = np.empty(shape=(n_kmers, nreads), dtype=np.float)
    mod_pos_list = []
    nreads_mod = 0

    # Fill in arrays with non modified data per position
    for pos in range(n_kmers):
        kmer_seq =  ref_seq[pos:pos+5]
        kmer_model = model_df.loc[kmer_seq]

        # Sample intensity
        intensity_array[pos] = get_valid_distr_data(
            loc = kmer_model["model_intensity_loc"],
            scale = kmer_model["model_intensity_scale"],
            min = None if not_bound else kmer_model["raw_intensity_min"],
            max = None if not_bound else kmer_model["raw_intensity_max"],
            sp_distrib = sp_logistic,
            size=nreads,
            data_rand_seed=data_rand_seed)

        # Sample dwell
        dwell_array[pos] = get_valid_distr_data(
            loc = kmer_model["model_dwell_loc"],
            scale = kmer_model["model_dwell_scale"],
            min = None if not_bound else kmer_model["raw_dwell_min"],
            max = None if not_bound else kmer_model["raw_dwell_max"],
            sp_distrib = sp_wald,
            size=nreads,
            data_rand_seed=data_rand_seed)

    # If modifications are required, edit the values for randomly picked positions + adjacent positions if a context was given
    if mod_reads_freq and mod_bases_freq:
        # Define number of reads to modify and number not to modify
        nreads_mod = int(np.rint(nreads*mod_reads_freq))
        # Define positions to modify base on mod_base_freq and mod_base_type
        mod_pos_list = find_valid_pos_list(ref_seq, mod_bases_type, mod_bases_freq, min_mod_dist, pos_rand_seed)
        # if the modification context has to be extended
        mod_dict = make_mod_dict(intensity_mod, dwell_mod, mod_extend_context)


        for pos in mod_pos_list:
            for i in range(-mod_extend_context, mod_extend_context+1):
                pos_extend = pos+i
                if 0 <= pos_extend < n_kmers:
                    kmer_seq = ref_seq[pos_extend:pos_extend+5]
                    kmer_model = model_df.loc[kmer_seq]

                    if intensity_mod:
                        intensity_array[pos_extend][0:nreads_mod] = get_valid_distr_data(
                            loc = kmer_model["model_intensity_loc"],
                            scale = kmer_model["model_intensity_scale"],
                            min = None if not_bound else kmer_model["raw_intensity_min"],
                            max = None if not_bound else kmer_model["raw_intensity_max"],
                            mod = kmer_model["model_intensity_std"]*mod_dict["intensity"][i],
                            sp_distrib = sp_logistic,
                            size = nreads_mod,
                            data_rand_seed = data_rand_seed)

                    if dwell_mod:
                        dwell_array[pos_extend][0:nreads_mod] = get_valid_distr_data(
                            loc = kmer_model["model_dwell_loc"],
                            scale = kmer_model["model_dwell_scale"],
                            min = None if not_bound else kmer_model["raw_dwell_min"],
                            max = None if not_bound else kmer_model["raw_dwell_max"],
                            mod = kmer_model["model_dwell_std"]*mod_dict["dwell"][i],
                            sp_distrib = sp_wald,
                            size = nreads_mod,
                            data_rand_seed = data_rand_seed)

    return (intensity_array, dwell_array, mod_pos_list, nreads_mod)

def find_valid_pos_list(ref_seq, mod_bases_type, mod_bases_freq, min_mod_dist, pos_rand_seed=42):
    """"""
    pos_list = []
    for i in range(len(ref_seq)-4):
        if ref_seq[i] == mod_bases_type:
            pos_list.append(i)
    n_samples = int(np.rint(len(pos_list)*mod_bases_freq))
    logger.debug("\tTry to find {} kmers to modify".format(n_samples))

    i = 0
    while True:
        np.random.seed(pos_rand_seed*i)
        a = np.random.choice(pos_list, n_samples, replace=False)
        a.sort()
        if np.ediff1d(a).min() >= min_mod_dist :
            logger.debug("\tFound valid combination for {} kmers".format(n_samples))
            logger.debug("\tmodified positions: {}".format(a))
            break
        if i == 1000:
            i = 0
            n_samples -= 1
        i+=1
    return a

def get_valid_distr_data(loc, scale, size, sp_distrib, mod=None, min=None, max=None, max_tries=10000, data_rand_seed=None):
    """"""

    # If a mofidier is given modify the loc, min and max value accordingly
    if mod:
        loc+=mod
        if min:
            min+=mod
        if max:
            max+=mod

    # Define lower and upper bound if not given
    if not min:
        min=0
    if not max:
        max=np.finfo(np.float64).max

    # Try to sample the required number of data point
    i = 0
    while True:
        iter_rand_seed = None if data_rand_seed is None else data_rand_seed + i
        np.random.seed(iter_rand_seed)
        data = sp_distrib.rvs(loc=loc, scale=scale, size=size, random_state=iter_rand_seed)
        if data.min() > min and data.max() < max:
            return data

        # Safety trigger
        i+=1
        # Fall back to safe bounds if min max bounds fails to yield valid values
        if i > max_tries:
            logger.debug("\tCould not find valid values with min max bounds. Fall back to safe bounds")
            logger.debug("\tYou should consider using the `not_bound` option")
            min=0
            max=np.finfo(np.float64).max
        # If too many tries raise an exception
        if i > max_tries*2:
            raise NanocomporeError("Could not find valid data after {} tries".format(i))

def make_mod_dict(intensity_mod, dwell_mod, mod_extend_context):
    """Compute a harmonic series per values depending on the context length"""
    pos_list = list(range(-mod_extend_context, mod_extend_context+1))
    d = OrderedDict()
    d["intensity"] = {i: intensity_mod*(1/(abs(i)+1)) for i in pos_list}
    d["dwell"] = {i: dwell_mod*(1/(abs(i)+1)) for i in pos_list}
    return d

def array_join(sep, array):
    s = ""
    for i in array:
        s+="{}{}".format(i,sep)
    return s[:-len(sep)]

def parse_mod_pos_file(path):
    """ Parses a pos file generated by SimReads()
    and returns a dict of lists where the keys are the ref_ids
    """
    with open(path) as f:
        pos = dict(line.strip().split('\t') for line in f)
    pos.pop('ref_id')
    for k,v in pos.items():
        pos[k]=[int(i) for i in v.split(';')]
    return(pos)



# -*- coding: utf-8 -*-

#~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~#
# Std lib
from collections import *
import logging
from loguru import logger
import random

# Third party
import numpy as np
from tqdm import tqdm
from pyfaidx import Fasta

# Local package
from nanocompore.common import *

# Set global random seed
downsample_random_seed = 42
random.seed(downsample_random_seed)

#~~~~~~~~~~~~~~MAIN CLASS~~~~~~~~~~~~~~#
class Whitelist(object):

    #~~~~~~~~~~~~~~MAGIC METHODS~~~~~~~~~~~~~~#
    def __init__(self,
        eventalign_fn_dict,
        fasta_fn,
        min_coverage = 10,
        min_ref_length = 100,
        downsample_high_coverage = False,
        max_invalid_kmers_freq = 0.1,
        max_NNNNN_freq = 0.1,
        max_mismatching_freq = 0.1,
        max_missing_freq = 0.1,
        select_ref_id = [],
        exclude_ref_id = []):
        """
        #########################################################
        * eventalign_fn_dict
            Multilevel dictionnary indicating the condition_label, sample_label and file name of the eventalign_collapse output
            example d = {"S1": {"R1":"path1.tsv", "R2":"path2.tsv"}, "S2": {"R1":"path3.tsv", "R2":"path4.tsv"}}
        * fasta_fn
            Path to a fasta file corresponding to the reference used for read alignemnt
        * min_coverage
            minimal coverage required in both samples
        * min_ref_length
            minimal length of a reference transcript to be considered in the analysis
        * downsample_high_coverage
            For reference with higher coverage, downsample by randomly selecting reads.
        * max_invalid_kmers_freq
            maximum frequency of NNNNN, mismatching and missing kmers in reads
            If None, then the max_NNNNN_freq, max_mismatching_freq, max_missing_freq agrs will be used instead
        * max_NNNNN_freq
            maximum frequency of NNNNN kmers in reads (1 to deactivate)
        * max_mismatching_freq
            maximum frequency of mismatching kmers in reads (1 to deactivate)
        * max_missing_freq
            maximum frequency of missing kmers in reads (1 to deactivate)
        * select_ref_id
            if given, only reference ids in the list will be selected for the analysis
        * exclude_ref_id
            if given, refid in the list will be excluded from the analysis
        """

        # Check index files
        self.__filter_invalid_kmers = True
        for sample_dict in eventalign_fn_dict.values():
            for fn in sample_dict.values():
                idx_fn = fn+".idx"
                if not access_file(idx_fn):
                    raise NanocomporeError("Cannot access eventalign_collapse index file {}".format(idx_fn))
                # Check header line and set a flag to skip filter if the index file does not contain kmer status information
                with open(idx_fn, "r") as fp:
                    header = fp.readline().rstrip().split("\t")
                if not all_values_in (("ref_id", "read_id", "byte_offset", "byte_len"), header):
                    raise NanocomporeError("The index file {} does not contain the require header fields".format(idx_fn))
                if not all_values_in (("kmers", "NNNNN_kmers", "mismatch_kmers", "missing_kmers"), header):
                    self.__filter_invalid_kmers = False
                    logger.debug("Invalid kmer information not available in index file")

        self.__eventalign_fn_dict = eventalign_fn_dict

        # Get number of samples
        n = 0
        for sample_dict in self.__eventalign_fn_dict.values():
            for sample_lab in sample_dict.keys():
                n+=1
        self.__n_samples = n

        # Test is Fasta can be opened
        try:
            with Fasta(fasta_fn):
                self._fasta_fn = fasta_fn
        except IOError:
            raise NanocomporeError("The fasta file cannot be opened")

        # Create reference index for both files
        logger.info("Reading eventalign index files")
        ref_reads = self.__read_eventalign_index(
            eventalign_fn_dict = eventalign_fn_dict,
            max_invalid_kmers_freq = max_invalid_kmers_freq,
            max_NNNNN_freq = max_NNNNN_freq,
            max_mismatching_freq = max_mismatching_freq,
            max_missing_freq = max_missing_freq,
            select_ref_id = select_ref_id,
            exclude_ref_id = exclude_ref_id)

        # Filtering at transcript level
        logger.info("Filtering out references with low coverage")
        self.ref_reads = self.__select_ref(
            ref_reads = ref_reads,
            min_coverage=min_coverage,
            min_ref_length=min_ref_length,
            downsample_high_coverage=downsample_high_coverage)

        self.__min_coverage = min_coverage
        self.__min_ref_length = min_ref_length
        self.__downsample_high_coverage = downsample_high_coverage
        self.__max_invalid_kmers_freq = max_invalid_kmers_freq

    def __repr__(self):
        return "Whitelist: Number of references: {}".format(len(self))

    def __str__(self):
        m = ""
        for ref_id, ref_dict in self.ref_reads.items():
            m += "{}\n".format(ref_id)
            for cond_lab, cond_dict in ref_dict.items():
                for sample_lab, read_list in cond_dict.items():
                    m += "\t{} {}\n".format(cond_lab, sample_lab)
                    for reads in read_list:
                        m += "\t\t"
                        for k,v in reads.items():
                            m += "{}:{}  ".format(k,v)
                        m += "\n"
        return m

    def __len__(self):
        return len(self.ref_reads)

    def __iter__(self):
        for i, j in self.ref_reads.items():
            yield(i,j)

    def __getitem__(self, items):
        return self.ref_reads.get(items, None)

    #~~~~~~~~~~~~~~PUBLIC METHODS AND PROPERTIES~~~~~~~~~~~~~~#
    @property
    def ref_id_list(self):
        return list(self.ref_reads.keys())

    #~~~~~~~~~~~~~~PRIVATE METHODS~~~~~~~~~~~~~~#
    def __read_eventalign_index(self,
        eventalign_fn_dict,
        max_invalid_kmers_freq,
        max_NNNNN_freq,
        max_mismatching_freq,
        max_missing_freq,
        select_ref_id,
        exclude_ref_id):
        """Read the 2 index files and sort by sample and ref_id in a multi level dict"""

        ref_reads = OrderedDict()

        for cond_lab, sample_dict in eventalign_fn_dict.items():
            for sample_lab, fn in sample_dict.items():
                idx_fn = fn+".idx"
                with open(idx_fn) as fp:

                    # Get column names from header
                    col_names = fp.readline().rstrip().split()
                    c = Counter()
                    for line in fp:
                        # Transform line to dict and cast str numbers to actual numbers
                        read = numeric_cast_dict(keys=col_names, values=line.rstrip().split("\t"))
                        read_is_valid = True
                        # Filter out ref_id if a select_ref_id list or exclude_ref_id list was provided
                        if select_ref_id and not read["ref_id"] in select_ref_id:
                            read_is_valid = False
                            c["Ref_id not in select list"]+=1
                            logger.trace("Ref_id not in select list")
                        elif exclude_ref_id and read["ref_id"] in exclude_ref_id:
                            read_is_valid = False
                            c["Ref_id in exclude list"]+=1
                            logger.trace("Ref_id in exclude list")

                        # Filter out reads with high number of invalid kmers if information available
                        if self.__filter_invalid_kmers:
                            if max_invalid_kmers_freq:
                                invalid_kmers_freq = (read["NNNNN_kmers"]+read["mismatch_kmers"]+read["missing_kmers"])/read["kmers"]
                                if invalid_kmers_freq > max_invalid_kmers_freq:
                                    read_is_valid = False
                                    c["High fraction of invalid kmers"]+=1
                                    logger.trace("High fraction of invalid kmers ({}%) for read {}".format(round(invalid_kmers_freq*100,2), read["read_id"]))
                            else:
                                NNNNN_kmers_freq = read["NNNNN_kmers"]/read["kmers"]
                                max_mismatching_freq = read["mismatch_kmers"]/read["kmers"]
                                max_missing_freq = read["missing_kmers"]/read["kmers"]
                                if NNNNN_kmers_freq > max_NNNNN_freq:
                                    read_is_valid = False
                                    c["High fraction of NNNNN kmers"]+=1
                                    logger.trace("High fraction of NNNNN kmers ({}%) for read {}".format(round(NNNNN_kmers_freq*100,2), read["read_id"]))
                                elif max_mismatching_freq > max_mismatching_freq:
                                    read_is_valid = False
                                    c["High fraction of mismatching kmers"]
                                    logger.trace("High fraction of mismatching kmers ({}%) for read {}".format(round(max_mismatching_freq*100,2), read["read_id"]))
                                elif max_missing_freq > max_missing_freq:
                                    read_is_valid = False
                                    c["High fraction of missing kmers"]+=1
                                    logger.trace("High fraction of missing kmers ({}%) for read {}".format(round(max_missing_freq*100,2), read["read_id"]))

                        if read_is_valid:
                            # Create dict arborescence and save valid reads
                            if not read["ref_id"] in ref_reads:
                                ref_reads[read["ref_id"]] = OrderedDict()
                            if not cond_lab in ref_reads[read["ref_id"]]:
                                ref_reads[read["ref_id"]][cond_lab] = OrderedDict()
                            if not sample_lab in ref_reads[read["ref_id"]][cond_lab]:
                                ref_reads[read["ref_id"]][cond_lab][sample_lab] = []

                            # Fill in list of reads
                            ref_reads[read["ref_id"]][cond_lab][sample_lab].append(read)
                            c ["valid reads"] += 1

                logger.debug("\tCondition:{} Sample:{} {}".format(cond_lab, sample_lab, counter_to_str(c)))
        # Fill in missing condition/sample slots in case
        # a ref_id is missing from one of the eventalign files
        for ref_id in ref_reads.keys():
            for cond_lab, sample_dict in eventalign_fn_dict.items():
                for sample_lab in sample_dict.keys():
                    if not cond_lab in ref_reads[ref_id]:
                        ref_reads[ref_id][cond_lab] = OrderedDict()
                    if not sample_lab in ref_reads[ref_id][cond_lab]:
                        ref_reads[ref_id][cond_lab][sample_lab] = []
        logger.info("\tReferences found in index: {}".format(len(ref_reads)))
        return ref_reads

    def __select_ref(self,
        ref_reads,
        min_coverage,
        min_ref_length,
        downsample_high_coverage):
        """Select ref_id with a minimal coverage in both sample + downsample if needed"""
        valid_ref_reads = OrderedDict()
        c = Counter()
        with Fasta(self._fasta_fn) as fasta:
            for ref_id, ref_dict in ref_reads.items():
                try:
                    # Discard reference transcripts shorter than the threshold
                    assert len(fasta[ref_id]) > min_ref_length
                    valid_dict = OrderedDict()
                    for cond_lab, cond_dict in ref_dict.items():
                        valid_dict[cond_lab] = OrderedDict()
                        for sample_lab, read_list in cond_dict.items():
                            logger.trace(f"Asserting if {ref_id} has enough coverage in {sample_lab}")
                            # Filter out if coverage too low
                            assert len(read_list) >= min_coverage
                            logger.trace(f"ref_id {ref_id} has {len(read_list)} reads in {sample_lab}")
                            # Downsample if coverage too high
                            if downsample_high_coverage and len(read_list) > downsample_high_coverage:
                                read_list = random.sample(read_list, downsample_high_coverage)
                            valid_dict[cond_lab][sample_lab]=read_list


                    # If all valid add to new dict
                    logger.trace(f"ref_id {ref_id} has enough coverage in all samples: keeping it")
                    valid_ref_reads [ref_id] = valid_dict

                    # Save extra info for debug
                    c["valid_ref_id"] += 1
                    for cond_lab, cond_dict in valid_dict.items():
                        for sample_lab, read_list in cond_dict.items():
                            lab = "{} {} Reads".format(cond_lab, sample_lab)
                            c[lab] += len(read_list)

                except AssertionError:
                    logger.trace(f"ref_id {ref_id} does not have enough coverage in at least one sample: discarding it")
                    c["invalid_ref_id"] += 1

        logger.debug(counter_to_str(c))
        logger.info("\tReferences remaining after reference coverage filtering: {}".format(len(valid_ref_reads)))
        return valid_ref_reads



# -*- coding: utf-8 -*-
# https://github.com/python-poetry/poetry/pull/2366#issuecomment-652418094
from nanocompore.common import get_version

__version__ = get_version()

__description__ = 'Software package that identifies raw signal changes between two conditions from https://github.com/jts/nanopolish resquiggled dRNA-Seq data.'



# -*- coding: utf-8 -*-

#~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~#
# Std lib
from loguru import logger
from collections import *
import shelve
from dbm import error as dbm_error
from math import log
import re
import sys
from pkg_resources import resource_filename
import itertools

# Third party
from pyfaidx import Fasta
import pandas as pd
import numpy as np
import matplotlib.pyplot as pl
import matplotlib.patches as mpatches
import seaborn as sns
from bedparse import bedline
from statsmodels.stats.multitest import multipletests
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import scale as scale

# Local package
from nanocompore.common import *

# Logger setup

#~~~~~~~~~~~~~~MAIN CLASS~~~~~~~~~~~~~~#
class SampCompDB(object):
    """ Wrapper over the result shelve SampComp """

    #~~~~~~~~~~~~~~FUNDAMENTAL METHODS~~~~~~~~~~~~~~#
    def __init__(self,
        db_fn:str,
        fasta_fn:str,
        bed_fn:str = None,
        run_type:str = "RNA"):
        """
        Import a shelve db and a fasta reference file. Automatically returned by SampComp
        Can also be manually created from an existing shelve db output
        * db_fn
            Path to a database file previously created with SampComp
        * fasta_fn
            Path to a fasta file corresponding to the reference used for read alignemnt
        * bed_fn
            Path to a BED file containing the annotation of the transcriptome used as reference when mapping
        * run_type
            Define the run type model to import {RNA, DNA}
        """

        logger.info("Loading SampCompDB")

        # Try to get ref_id list and metadata from shelve db
        try:
            with shelve.open(db_fn, flag='r') as db:
                # Try to get metadata from db
                try:
                    logger.debug("\tReading Metadata")
                    self._metadata = db['__metadata']
                except KeyError:
                    raise NanocomporeError("The result database does not contain metadata")
                # Try to load read_ids
                try:
                    logger.debug("\tLoading list of reference ids")
                    self.ref_id_list = db['__ref_id_list']
                except KeyError:
                    logger.debug("\tCannot find the ref_id_list in shelve. Try to build the list from entries")
                    self.ref_id_list = [k for k in db.keys() if k not in  ['__metadata', '__ref_id_list']]
                if not self.ref_id_list:
                    logger.info("The result database is empty")
                    return None
        except dbm_error:
            raise NanocomporeError("The result database cannot be opened")

        # Save db prefix and db path
        self._db_fn = db_fn

        logger.debug("Checking files and arg values")
        # Check if fasta and bed files exist
        if not access_file(fasta_fn):
            raise NanocomporeError("{} is not a valid FASTA file".format(fasta_fn))
        self._fasta_fn = fasta_fn
        if bed_fn and not access_file(bed_fn):
            raise NanocomporeError("{} is not a valid BED file".format(bed_fn))
        self._bed_fn = bed_fn

        # Define model depending on run_type
        if run_type == "RNA":
            model_fn = resource_filename("nanocompore", "models/kmers_model_RNA_r9.4_180mv.tsv")
            self._model_df = pd.read_csv(model_fn, sep="\t", comment="#", index_col=0)
        else:
            raise NanocomporeError("Only RNA is implemented at the moment")

        # Create results df with adjusted p-values
        if self._metadata["pvalue_tests"]:
            logger.info("Calculate results")
            self.results = self.__calculate_results(adjust=True)

    def __repr__(self):
        """readable description of the object"""
        s = "[{}]\n".format(self.__class__.__name__)
        for i,j in self._metadata.items():
            s+= "\t{}: {}\n".format(i,j)
        s+= "\tNumber of references: {}\n".format(len(self))
        return s

    #~~~~~~~~~~~~~~MAGIC METHODS~~~~~~~~~~~~~~#
    def __len__(self):
        return len(self.ref_id_list)

    def __iter__(self):
        with shelve.open(self._db_fn, flag = "r") as db:
            for k, v in db.items():
                if not k in  ['__metadata', '__ref_id_list']:
                    yield(k, v)

    def __getitem__(self, items):
        with shelve.open(self._db_fn, flag = "r") as db:
            if items in db:
                return db[items]
            else:
                raise KeyError("Item not found in the database")

    #~~~~~~~~~~~~~~PRIVATE  METHODS~~~~~~~~~~~~~~#

    def __calculate_results(self, adjust=True):
        """"""
        # Collect all pvalue results in a dataframe
        l = []
        for ref_id, ref_pos_list in self:
            for pos, pos_dict in enumerate(ref_pos_list):
                if "txComp" in pos_dict:
                    row_dict = OrderedDict()
                    row_dict["ref_id"] = ref_id
                    row_dict["pos"] = pos
                    row_dict["ref_kmer"] = pos_dict["ref_kmer"]
                    for test in self._metadata["pvalue_tests"]:
                        if test in pos_dict["txComp"]:
                            pval = pos_dict["txComp"][test]
                            row_dict[test] = pval
                    l.append(row_dict)
        df = pd.DataFrame(l)

        if self._bed_fn:
            bed_annot={}
            try:
                with open(self._bed_fn) as tsvfile:
                    for line in tsvfile:
                        record_name=line.split('\t')[3]
                        if( record_name in self.ref_id_list):
                            bed_annot[record_name]=bedline(line.split('\t'))
            except:
                raise NanocomporeError("Can't open BED file")
            if len(bed_annot) != len(self.ref_id_list):
                raise NanocomporeError("Some references are missing from the BED file provided")

            df['genomicPos'] = df.apply(lambda row: bed_annot[row['ref_id']].tx2genome(coord=row['pos'], stranded=True),axis=1)
            # This is very inefficient. We should get chr and strand only once per transcript, ideally when writing the BED file
            df['chr'] = df.apply(lambda row: bed_annot[row['ref_id']].chr,axis=1)
            df['strand'] = df.apply(lambda row: bed_annot[row['ref_id']].strand,axis=1)
            df=df[['ref_id', 'pos', 'chr', 'strand', 'genomicPos', 'ref_kmer']+self._metadata["pvalue_tests"]]
        else:
            df=df[['ref_id', 'pos', 'ref_kmer']+self._metadata["pvalue_tests"]]

        if adjust:
            for col in self._metadata["pvalue_tests"]:
                df[col] = self.__multipletests_filter_nan(df[col], method="fdr_bh")
        return df

    def __get_kmer_list(self, ref_id, start, end, kmer_size=5):
        """ Extract fasta record corresponding to ref with error handling """
        try:
            with Fasta(self._fasta_fn) as fasta:
                fasta =(fasta [ref_id])
                seq = str(fasta[start:end+5])
                kmer_list = []
                for i in range(end-start):
                    kmer_list.append(seq[i:i+5])
                return kmer_list
        except KeyError:
            raise NanocomporeError("Reference id not present in fasta file")

    def __get_positions(self, ref_id, start=None, end=None):
        """ Verify start and end and if not available try to infer them"""
        try:
            with Fasta(self._fasta_fn) as fasta:
                max_len = len(fasta [ref_id])-4
        except KeyError:
            raise NanocomporeError("Reference id not present in fasta file")
        if not start or start < 0:
            start = 0
        if not end or end > max_len:
            end = max_len
        if start > end:
            raise NanocomporeError("End coordinate has to be higher or equal to start")
        return(start, end)

    @staticmethod
    def __color_generator(palette, n):
        if type(palette) in (list, tuple, set):
            for col in itertools.islice(itertools.cycle(palette), n):
                yield(col)
        elif type(palette) == str:
            palette = sns.mpl_palette(palette, n)
            for i in range(n):
                yield(palette[i])
        else:
            raise NanocomporeError ("Invalid palette type")

    @staticmethod
    def __multipletests_filter_nan(pvalues, method="fdr_bh"):
        """
        Performs p-value correction for multiple hypothesis testing
        using the method specified. The pvalues list can contain
        np.nan values, which are ignored during p-value correction.
        test: input=[0.1, 0.01, np.nan, 0.01, 0.5, 0.4, 0.01, 0.001, np.nan, np.nan, 0.01, np.nan]
        out: array([0.13333333, 0.016     ,        nan, 0.016     , 0.5       ,
        0.45714286, 0.016     , 0.008     ,        nan,        nan,
        0.016     ,        nan])
        """
        if all([np.isnan(p) for p in pvalues]):
            return pvalues

        pvalues_no_nan = [p for p in pvalues if not np.isnan(p)]
        corrected_p_values = multipletests(pvalues_no_nan, method=method)[1]
        for i, p in enumerate(pvalues):
            if np.isnan(p):
                corrected_p_values=np.insert(corrected_p_values, i, np.nan, axis=0)
        return(corrected_p_values)

    #~~~~~~~~~~~~~~PUBLIC METHODS~~~~~~~~~~~~~~#
    def save_all (self, outpath_prefix=None, pvalue_thr=0.01):
        """
        Save all text reports including genomic coordinate if a bed file was provided
        * outpath_prefix
            outpath + prefix to use as a basename for output files.
            If not given, it will use the same prefix as the database.
        * pvalue_thr
            pvalue threshold to report significant sites in bed files
        """
        if not outpath_prefix:
            outpath_prefix = self._db_fn.replace("SampComp.db", "")
        logger.debug("Save reports to {}".format(outpath_prefix))

        # Save reports
        logger.debug("\tSaving extended tabular report")
        self.save_report(output_fn = outpath_prefix+"nanocompore_results.tsv")
        logger.debug("\tSaving shift results")
        self.save_shift_stats(output_fn = outpath_prefix+"nanocompore_shift_stats.tsv")

        # Save bed and bedgraph files for each method used
        if self._bed_fn:
            logger.debug("\tSaving significant genomic coordinates in Bed and Bedgraph format")
            for m in self._metadata["pvalue_tests"]:
                self.save_to_bed(
                    output_fn = outpath_prefix+"sig_sites_{}_thr_{}.bed".format(m, pvalue_thr),
                    bedgraph=False, pvalue_field=m, pvalue_thr=pvalue_thr, span=5, title="Nanocompore Significant Sites")
                self.save_to_bed(
                    output_fn = outpath_prefix+"sig_sites_{}_thr_{}.bedgraph".format(m, pvalue_thr),
                    bedgraph=True, pvalue_field=m, pvalue_thr=pvalue_thr, title="Nanocompore Significant Sites")

    def save_to_bed(self, output_fn=None, bedgraph=False, pvalue_field=None, pvalue_thr=0.01, span=5, convert=None, assembly=None, title=None):
        """
        Save the position of significant positions in the genome space in BED6 or BEDGRAPH format.
        The resulting file can be used in a genome browser to visualise significant genomic locations.
        The option is only available if `SampCompDB` if initialised with a BED file containing genome annotations.
        * output_fn
            Path to file where to write the data
        * bedgraph
            save file in bedgraph format instead of bed
        * pvalue_field
            specifies what column to use as BED score (field 5, as -log10)
        * pvalue_thr
            only report positions with pvalue<=thr
        * span
            The size of each BED feature.
            If size=5 (default) features correspond to kmers.
            If size=1 features correspond to the first base of each kmer.
        * convert
            one of 'ensembl_to_ucsc' or 'ucsc_to_ensembl". Convert chromosome named between Ensembl and Ucsc conventions
        * assembly
            required if convert is used. One of "hg38" or "mm10"
        """
        if self._bed_fn is None:
            raise NanocomporeError("In order to generate a BED file SampCompDB needs to be initialised with a transcriptome BED")
        if span < 1:
            raise NanocomporeError("span has to be >=1")
        if span != 5 and bedgraph:
            raise NanocomporeError("Span is ignored when generating bedGraph files")
        if pvalue_field not in self.results:
            raise NanocomporeError(("The field '%s' is not in the results" % pvalue_field))
        if "results" not in self.__dict__:
            raise NanocomporeError("It looks like there's not results slot in SampCompDB")
        if convert not in [None, "ensembl_to_ucsc", "ucsc_to_ensembl"]:
            raise NanocomporeError("Convert value not valid")
        if convert is not None and assembly is None:
            raise NanocomporeError("The assembly argument is required in order to do the conversion. Choose one of 'hg38' or 'mm10' ")

        with open(output_fn, "w") as bed_file:
            if title is not None:
                if not bedgraph:
                    bed_file.write('track type=bed name="%s" description="%s"\n'%(title,pvalue_field))
                else:
                    bed_file.write('track type=bedGraph name="%s" description="%s"\n'%(title,pvalue_field))

            Record = namedtuple('Record', ['chr', 'genomicPos', 'ref_id','strand', 'ref_kmer', pvalue_field ])
            for record in self.results[ list(Record._fields) ].itertuples(index=False, name="Record"):
                pvalue = getattr(record, pvalue_field)
                if np.isnan(pvalue):
                    pvalue=0
                elif pvalue < sys.float_info.min:
                    pvalue = -log(sys.float_info.min, 10)
                else:
                    pvalue=-log(pvalue, 10)
                if not bedgraph and pvalue >= -log(pvalue_thr, 10):
                    if record.strand == "+":
                        line=bedline([record.chr, record.genomicPos, record.genomicPos+span, "%s_%s" % (record.ref_id, record.ref_kmer), pvalue, record.strand])
                    else:
                        line=bedline([record.chr, record.genomicPos-(span-1), record.genomicPos+1, "%s_%s" % (record.ref_id, record.ref_kmer), pvalue, record.strand])

                    if convert == "ensembl_to_ucsc":
                        line=line.translateChr(assembly=assembly, target="ucsc", patches=True)
                    elif convert == "ucsc_to_ensembl":
                        line=line.translateChr(assembly=assembly, target="ens", patches=True)
                    bed_file.write("%s\t%s\t%s\t%s\t%s\t%s\n" % (line.chr, line.start, line.end, line.name, line.score, line.strand))
                elif bedgraph:
                    if record.strand == "+":
                        line=bedline([record.chr, record.genomicPos+2, record.genomicPos+3, "%s_%s" % (record.ref_id, record.ref_kmer), pvalue, record.strand])
                    else:
                        line=bedline([record.chr, record.genomicPos-2, record.genomicPos-1, "%s_%s" % (record.ref_id, record.ref_kmer), pvalue, record.strand])
                    if convert == "ensembl_to_ucsc":
                        line=line.translateChr(assembly=assembly, target="ucsc", patches=True)
                    elif convert == "ucsc_to_ensembl":
                        line=line.translateChr(assembly=assembly, target="ens", patches=True)
                    bed_file.write("%s\t%s\t%s\t%s\n" % (line.chr, line.start, line.end, line.score))

    def save_report(self, output_fn:str=None):
        """
        Saves a tabulated text dump of the database containing all the statistical results for all the positions
        * output_fn
            Path to file where to write the data. If None, data is returned to the standard output.
        """
        if output_fn is None:
            fp = sys.stdout
        elif isinstance(output_fn, str):
            try:
                fp = open(output_fn, "w")
            except:
                raise NanocomporeError("Error opening output file %s"%output_fn)
        else:
            raise NanocomporeError("output_fn needs to be a string or None")

        headers = ['pos', 'chr', 'genomicPos', 'ref_id', 'strand', 'ref_kmer']+self._metadata["pvalue_tests"]

        # Read extra GMM info from the shelve
        if "GMM" in self._metadata["comparison_methods"]:
            headers.extend(["GMM_cov_type", "GMM_n_clust", "cluster_counts"])
            # Conditional add if logit or Anova
            if "GMM_anova_pvalue" in self._metadata["pvalue_tests"]:
                headers.append("Anova_delta_logit")
            if "GMM_logit_pvalue" in self._metadata["pvalue_tests"]:
                headers.append("Logit_LOR")

        # Write headers to file
        fp.write('\t'.join([ str(i) for i in headers ])+'\n')

        # We loop over the IDs so that ref_pos_list can be prefetched for each transcript
        for cur_id in self.ref_id_list:
            cur_ref_pos_list = self[cur_id]
            for record in self.results[self.results.ref_id == cur_id ].itertuples():
                if "GMM" in self._metadata["comparison_methods"]:
                    record_txComp = cur_ref_pos_list[record.pos]['txComp']
                line = []
                for f in headers:
                    if f in record._fields:
                        line.append(getattr(record, f))
                    elif f == "GMM_cov_type":
                        line.append(record_txComp['GMM_model']['model'].covariance_type)
                    elif f == "GMM_n_clust":
                        line.append(record_txComp['GMM_model']['model'].n_components)
                    elif f == "cluster_counts":
                        line.append(record_txComp['GMM_model']['cluster_counts'])
                    elif f == "Anova_delta_logit":
                        line.append(record_txComp['GMM_anova_model']['delta_logit'])
                    elif f == "Logit_LOR":
                        line.append(record_txComp['GMM_logit_model']['coef'])
                    else: line.append("NA")
                fp.write('\t'.join([ str(i) for i in line ])+'\n')
        fp.close()

    def save_shift_stats(self, output_fn=None):
        """
        Save the mean, median and sd intensity and dwell time for each condition and for each position.
        This can be used to evaluate the intensity of the shift for significant positions.
        * output_fn
            Path to file where to write the data. If None, data is returned to the standard output.
        """
        if output_fn is None:
            fp = sys.stdout
        elif isinstance(output_fn, str):
            try:
                fp = open(output_fn, "w")
            except:
                raise NanocomporeError("Error opening output file %s"%output_fn)
        else:
            raise NanocomporeError("output_fn needs to be a string or None")


        headers = ['c1_mean_intensity', 'c2_mean_intensity', 'c1_median_intensity', 'c2_median_intensity', 'c1_sd_intensity', 'c2_sd_intensity', 'c1_mean_dwell', 'c2_mean_dwell', 'c1_median_dwell', 'c2_median_dwell', 'c1_sd_dwell', 'c2_sd_dwell']
        fp.write('\t'.join([ str(i) for i in ["ref_id", "pos"]+headers ])+'\n')
        for tx, refpos in self:
            for pos, refpos_list in enumerate(refpos):
                if "txComp" in refpos_list:
                    ss = refpos_list['txComp']['shift_stats']
                    if list(ss.keys()) != headers:
                        raise NanocomporeError("Mismatch in shift_stats headers")
                    line = [tx, pos, *ss.values()]
                    fp.write('\t'.join([ str(i) for i in line ])+'\n')
        fp.close()

    def list_significant_positions(self, ref_id=None, test=None, thr=0.05):
        """
        Return a list of positions of ref_id with p-value <= threshold for
        statistical test specified
        * ref_id
            Valid reference id name in the database
        * test
            Name of the test of interest
        * thr
            p-value threshold
        """
        if not test in self._metadata["pvalue_tests"]:
            raise NanocomporeError("The test requested ({}) does not exist".format(test))
        if not ref_id in self.ref_id_list:
            raise NanocomporeError("The reference requested ({}) is not in the DB".format(ref_id))
        sig = list(self.results[(self.results['ref_id'] == ref_id) & (self.results[test] <= thr)]['pos'])
        return(sig)

    # def list_most_significant_references(self, n=10):
    #     pass

    #~~~~~~~~~~~~~~PLOTTING METHODS~~~~~~~~~~~~~~#
    def plot_pvalue( self,
        ref_id:str,
        start:int=None,
        end:int=None,
        kind:str="lineplot",
        threshold:float=0.01,
        figsize:tuple=(30,10),
        palette:str="Set2",
        plot_style:str="ggplot",
        tests:str=None):
        """
        Plot pvalues per position (by default plot all fields starting by "pvalue")
        * ref_id
            Valid reference id name in the database
        * start
            Start coordinate
        * end
            End coordinate (included)
        * kind
            kind of plot to represent the data. {lineplot,barplot}
        * figsize
            Length and heigh of the output plot
        * palette
            Colormap. See https://matplotlib.org/users/colormaps.html, https://matplotlib.org/examples/color/named_colors.html
        * plot_style
            Matplotlib plotting style. See https://matplotlib.org/users/style_sheets.html
        * tests
            Limit the pvalue methods shown in the plot. Either a list of methods or a string coresponding to a part of the name
        """
        # Extract fasta and positions
        start, end = self.__get_positions(ref_id, start, end)
        kmer_list = self.__get_kmer_list(ref_id, start, end)
        pos_list = list(range(start, end))

        # Extract data from result dataframe
        df = self.results.query("ref_id==@ref_id and @start<=pos<@end")
        if df.empty:
            raise NanocomporeError("No data available for the selected interval")

        # list tests methods
        if not tests:
            tests = self._metadata["pvalue_tests"]
        else:
            if isinstance(tests, str):
                tests = tests.split(",")
            elif not isinstance(tests, list):
                raise NanocomporeError("Method must be either a string or a list")
            tests_set = set()
            for test in tests:
                for available_test in self._metadata["pvalue_tests"]:
                    if test in available_test:
                        tests_set.add(available_test)
            if not tests_set:
                raise NanocomporeError("No matching tests found in the list of available tests: {}".format(self._metadata["pvalue_tests"]))
            tests = sorted(list(tests_set))

        # Collect data for interval in a numpy array
        array = np.zeros(end-start, dtype=[(test, np.float) for test in tests])
        for id, row in df.iterrows():
            for test in tests:
                if np.isnan(row[test]):
                    v = 0
                else:
                    v = -np.log10(row[test])
                array[row["pos"]-start][test] = v
        # Cast collected results to dataframe
        df = pd.DataFrame(array, index=pos_list)

        # Define plotting style
        with pl.style.context(plot_style):
            fig, ax = pl.subplots(figsize=figsize)
            if kind == "lineplot":
                _ = sns.lineplot(data=df, palette=palette, ax=ax, dashes=False)
                if end-start<30:
                    _ = ax.set_xticks(pos_list)
                    _ = ax.set_xticklabels([ "{}\n{}".format(i, j) for i,j in zip(pos_list, kmer_list)])

            elif kind == "barplot":
                df = df.reset_index()
                df = df.melt(id_vars="index", var_name="method", value_name="pvalue")
                _ = sns.barplot(x="index", y="pvalue", hue="method", data=df, palette=palette, ax=ax)
                if end-start<30:
                    _ = ax.set_xticklabels([ "{}\n{}".format(i, j) for i,j in zip(pos_list, kmer_list)])
                else:
                    breaks = list(range(start, end+1, (end-start)//10))
                    _ = ax.set_xticklabels([ i if i in breaks else "" for i in pos_list])

            _ = ax.axhline(y=-np.log10(threshold), color="grey", linestyle=":", label="pvalue={}".format(threshold))
            _ = ax.legend(bbox_to_anchor=(1, 1), loc=2, facecolor="white", frameon=False)
            _ = ax.set_ylabel("-log (pvalue)")
            _ = ax.set_xlabel("Reference position")
            _ = ax.set_title(ref_id)
            pl.tight_layout()
            return(fig, ax)

    def plot_signal(self,
        ref_id:str,
        start:int=None,
        end:int=None,
        kind:str="violinplot",
        split_samples:bool=False,
        figsize:tuple=(30,10),
        palette:str="Set2",
        plot_style:str="ggplot"):
        """
        Plot the dwell time and median intensity distribution position per position
        Pointless for more than 50 positions at once as it becomes hard to distinguish
        * ref_id
            Valid reference id name in the database
        * start
            Start coordinate
        * end
            End coordinate (included)
        * kind
            Kind of plot {violinplot, boxenplot, swarmplot}
        * split_samples
            If samples for a same condition are represented separatly. If false they are merged per condition
        * figsize
            Length and heigh of the output plot
        * palette
            Colormap. See https://matplotlib.org/users/colormaps.html, https://matplotlib.org/examples/color/named_colors.html
        * plot_style
            Matplotlib plotting style. See https://matplotlib.org/users/style_sheets.html
        """

        # Extract data for ref_id
        ref_data = self[ref_id]
        start, end = self.__get_positions(ref_id, start, end)

        # Parse line position per position
        l_intensity = []
        l_dwell = []
        x_ticks_list = []
        model_intensity_list = []
        model_dwell_list = []

        # Extract data from database if position in db
        for pos in np.arange(start, end):
            ref_kmer=ref_data[pos]['ref_kmer']
            x_ticks_list.append("{}\n{}".format(pos, ref_kmer))
            model_intensity_list.append(self._model_df.loc[ref_kmer]["model_intensity_median"])
            model_dwell_list.append(np.log10(self._model_df.loc[ref_kmer]["model_dwell_median"]))

            for cond_lab, cond_dict in ref_data[pos]['data'].items():
                for sample_lab, sample_val in cond_dict.items():
                    lab = "{}_{}".format(cond_lab, sample_lab) if split_samples else cond_lab

                    # Add intensity and dwell values to list for curent pos / lab
                    if not sample_val["intensity"]:
                        l_intensity.append((pos, lab, None))
                    for value in sample_val["intensity"]:
                        l_intensity.append((pos, lab, value))
                    if not sample_val["dwell"]:
                        l_dwell.append((pos, lab, None))
                    for value in sample_val["dwell"]:
                        l_dwell.append((pos, lab, np.log10(value)))

        # Define ploting style
        with pl.style.context(plot_style):
            fig, (ax1, ax2) = pl.subplots(2,1, figsize=figsize, sharex=True)

            for ax, l, model in ((ax1,l_intensity,model_intensity_list), (ax2,l_dwell,model_dwell_list)):
                # Plot values
                df = pd.DataFrame(l, columns=["pos", "lab", "value"])
                if kind == "violinplot":
                    _ = sns.violinplot(x="pos", y="value", hue="lab", data=df, ax=ax, split=not split_samples, inner="quartile",
                                        bw=0.25, linewidth=1, scale="area", palette=palette, zorder=0)
                elif kind == "boxenplot":
                    _ = sns.boxenplot(x="pos", y="value", hue="lab", data=df, ax=ax, scale="area", palette=palette, zorder=0)
                elif kind == "swarmplot":
                    _ = sns.swarmplot(x="pos", y="value", hue="lab", data=df, ax=ax, dodge=True, palette=palette, zorder=0)
                else:
                    raise NanocomporeError("Not a valid plot kind {}".format(kind))
                # Plot model
                _ = ax.plot(model, color="black", marker="x", label="Model Median", linestyle="", zorder=1)

            # Adjust display
            _ = ax1.set_xlabel("")
            _ = ax2.set_xlabel("Reference position")
            _ = ax1.set_ylabel("Mean Intensity")
            _ = ax2.set_ylabel("log10 (Dwell Time)")
            _ = ax1.legend(bbox_to_anchor=(1, 1), loc=2, facecolor="white", frameon=False)
            _ = ax2.legend("")
            _ = ax2.set_xlim(-1, end-start)
            _ = ax2.set_xticklabels(x_ticks_list)
            _ = fig.suptitle("Reference:{}  Start:{}  End:{}".format(ref_id, start, end), y=1.01, fontsize=18)

            pl.tight_layout()
            return(fig, (ax1, ax2))

    def plot_coverage(self,
        ref_id:str,
        start:int=None,
        end:int=None,
        scale:bool=False,
        split_samples:bool=False,
        figsize:tuple=(30,5),
        palette:str="Set2",
        plot_style:str="ggplot"):
        """
        Plot the read coverage over a reference for all samples analysed
        * ref_id
            Valid reference id name in the database
        * start
            Start coordinate
        * end
            End coordinate (included)
        * figsize
            Length and heigh of the output plot
        * palette
            Colormap. See https://matplotlib.org/users/colormaps.html, https://matplotlib.org/examples/color/named_colors.html
        * plot_style
            Matplotlib plotting style. See https://matplotlib.org/users/style_sheets.html
        """
        # Extract data for ref_id
        ref_data = self[ref_id]
        start, end = self.__get_positions(ref_id, start, end)

        # Parse data from database
        l = []
        for pos in np.arange(start, end):
            for cond_lab, cond_dict in ref_data[pos]['data'].items():
                for sample_lab, sample_val in cond_dict.items():
                    l.append((pos, "{}_{}".format(cond_lab, sample_lab), sample_val["coverage"]))

        # Cast collected results to dataframe
        df = pd.DataFrame(l, columns=["pos", "sample", "cov"])
        if scale:
            df['cov'] = df.groupby('sample')['cov'].apply(lambda x: x/max(x))

        # Define plotting style
        with pl.style.context(plot_style):
            fig, ax = pl.subplots(figsize=figsize)
            _ = sns.lineplot( x="pos", y="cov", hue="sample", data=df, ax=ax, palette=palette, drawstyle="steps")
            if not scale:
                _ = ax.axhline(y=self._metadata["min_coverage"], linestyle=":", color="grey", label="minimal coverage")

            _ = ax.set_ylim(0, None)
            _ = ax.set_xlim(start, end-1)
            _ = ax.set_title("Reference:{}  Start:{}  End:{}".format(ref_id, start, end), fontsize=18)
            _ = ax.set_ylabel("Coverage")
            _ = ax.set_xlabel("Reference position")
            _ = ax.legend(bbox_to_anchor=(1, 1), loc=2, facecolor="white", frameon=False)

            pl.tight_layout()
            return(fig, ax)

    def plot_bleeding_hulk(self, ref_id, start=None, end=None, split_samples=False, figsize=(30,10)):
        self.plot_kmers_stats(ref_id, start, end, split_samples, figsize, "Accent")

    def plot_kmers_stats(self,
        ref_id:str,
        start:int=None,
        end:int=None,
        split_samples:bool=False,
        figsize:tuple=(30,10),
        palette:str="Accent",
        plot_style:str="ggplot"):
        """
        Fancy version of `plot_coverage` that also report missing, mismatching and undefined kmers status from Nanopolish
        * ref_id
            Valid reference id name in the database
        * start
            Start coordinate
        * end
            End coordinate (included)
        * figsize
            Length and heigh of the output plot
        * palette
            Colormap. See https://matplotlib.org/users/colormaps.html, https://matplotlib.org/examples/color/named_colors.html
        * plot_style
            Matplotlib plotting style. See https://matplotlib.org/users/style_sheets.html
        """
        # Extract data for ref_id
        ref_data = self[ref_id]
        start, end = self.__get_positions(ref_id, start, end)

        # Parse data from database
        d = OrderedDict()
        for pos in np.arange(start, end):
            for cond_lab, cond_dict in ref_data[pos]["data"].items():
                for samp_lab, sample_val in cond_dict.items():
                    lab = "{}_{}".format(cond_lab, samp_lab) if split_samples else cond_lab

                    # Create dict arborescence
                    if not lab in d:
                        d[lab] = OrderedDict()
                    if not pos in d[lab]:
                        d[lab][pos] = {"valid":0,"NNNNN":0,"mismatching":0,"missing":0}

                    # Fill-in with values
                    d[lab][pos]["valid"] += sample_val["kmers_stats"]["valid"]
                    d[lab][pos]["NNNNN"] += sample_val["kmers_stats"]["NNNNN"]
                    d[lab][pos]["mismatching"] += sample_val["kmers_stats"]["mismatching"]
                    d[lab][pos]["missing"] += sample_val["kmers_stats"]["missing"]

        with pl.style.context(plot_style):
            fig, axes = pl.subplots(len(d),1, figsize=figsize)
            for ax, (lab, pos_dict) in zip(axes, d.items()):
                sample_df = pd.DataFrame.from_dict(pos_dict, orient="index")

                _ = sample_df.plot.area(ax=ax, colormap=palette, legend=False)
                _ = ax.set_title(lab)
                _ = ax.set_ylabel("Coverage")
                _ = ax.set_xlim(start, end-1)

            _ = axes[-1].set_xlabel("Reference position")
            _ = axes[0].legend(bbox_to_anchor=(1, 1), loc=2, facecolor="white", frameon=False)
            _ = fig.suptitle("Reference:{}  Start:{}  End:{}".format(ref_id, start, end), y=1.02, fontsize=18)
            pl.tight_layout()

        return(fig, axes)

    def plot_position(self,
        ref_id:str,
        pos:int=None,
        split_samples:bool=False,
        figsize:tuple=(30,10),
        palette:str="Set2",
        plot_style:str="ggplot",
        xlim:tuple=(None,None),
        ylim:tuple=(None,None),
        alpha:float=0.3,
        pointSize:int=20,
        scatter:bool=True,
        kde:bool=True,
        model:bool=False,
        gmm_levels:int=50):
        """
        Plot the dwell time and median intensity at the given position as a scatter plot.
        * ref_id
            Valid reference id name in the database
        * pos
            Position of interest
        * split_samples
            If True, samples for a same condition are represented separately. If False, they are merged per condition
        * figsize
            Length and heigh of the output plot
        * palette
            Colormap. See https://matplotlib.org/users/colormaps.html, https://matplotlib.org/examples/color/named_colors.html
        * plot_style
            Matplotlib plotting style. See https://matplotlib.org/users/style_sheets.html
        * xlim
            A tuple of explicit limits for the x axis
        * ylim
            A tuple of explicit limits for the y axis
        * kde
            plot the KDE of the intensity/dwell bivarariate distributions in the two samples
        * scatter
            if True, plot the individual data points
        * pointSize
            int specifying the point size for the scatter plot
        * model
            If true, plot the GMM density estimate
        * gmm_levels
            number of contour lines to use for the GMM countour plot
        """
        # Extract data for ref_id
        ref_data = self[ref_id]

        # Check that position is valid
        if not isinstance(pos, int):
            raise NanocomporeError("pos must be a single position")
        if pos > len(ref_data):
            raise NanocomporeError("Position out of range")
        # if not ref_data[pos]['data']["intensity"] or not ref_data[pos]['data']["dwell"]:
        #     raise NanocomporeError("No data found for selected position")

        # Extract data from database if position in db
        ref_kmer = ref_data[pos]['ref_kmer']
        data = ref_data[pos]['data']

        # Sample colors in palette
        col_gen = self.__color_generator(palette=palette, n=self._metadata["n_samples"] if split_samples else 2)

        # Collect and transform data in dict
        plot_data_dict = OrderedDict()
        for cond_lab, cond_dict in ref_data[pos]["data"].items():
            if split_samples:
                for samp_lab, sample_val in cond_dict.items():
                    plot_data_dict["{}_{}".format(cond_lab, samp_lab)] = {
                        "intensity":scale(sample_val["intensity"]),
                        "dwell":scale(np.log10(sample_val["dwell"])),
                        "color":next(col_gen)}
            else:
                intensity_list = []
                dwell_list = []
                for samp_lab, sample_val in cond_dict.items():
                    intensity_list.append(sample_val["intensity"])
                    dwell_list.append(sample_val["dwell"])
                plot_data_dict[cond_lab] = {
                    "intensity":scale(np.concatenate(intensity_list)),
                    "dwell":scale(np.log10(np.concatenate(dwell_list))),
                    "color":next(col_gen)}

        # Add GMM model if required and available
        if model and 'txComp' in ref_data[pos] and 'GMM_model' in ref_data[pos]['txComp']:
            model = ref_data[pos]['txComp']['GMM_model']['model']
            if not isinstance(model, GaussianMixture):
                raise NanocomporeError("The GMM_model slot for this position is not an instance of the GaussianMixture class")
            condition_labels = tuple(data.keys())
            global_intensity = scale(np.concatenate(([v['intensity'] for v in data[condition_labels[0]].values()]+[v['intensity'] for v in data[condition_labels[1]].values()]), axis=None))
            global_dwell = scale(np.log10(np.concatenate(([v['dwell'] for v in data[condition_labels[0]].values()]+[v['dwell'] for v in data[condition_labels[1]].values()]), axis=None)))
            x = np.linspace(min(global_intensity), max(global_intensity), num=1000)
            y = np.linspace(min(global_dwell), max(global_dwell), num=1000)
            X, Y = np.meshgrid(x, y)
            XX = np.array([X.ravel(), Y.ravel()]).T
            Z = -model.score_samples(XX)
            Z = Z.reshape(X.shape)
        else:
            model = None

        # plot collected data
        with pl.style.context(plot_style):
            fig, ax = pl.subplots(figsize=figsize)

            for label, d in plot_data_dict.items():
                if kde:
                    _ = sns.kdeplot(
                        data=d["intensity"],
                        data2=d["dwell"],
                        cmap=sns.light_palette(d["color"], as_cmap=True),
                        ax=ax,
                        clip=((min(d["intensity"]), max(d["intensity"])), (min(d["dwell"]),max(d["dwell"]))))
                if scatter:
                    _ = ax.scatter(
                        x=d["intensity"],
                        y=d["dwell"],
                        color=d["color"],
                        label=label,
                        alpha=alpha,
                        s=pointSize)
            if model:
                _ = ax.contour(X, Y, Z, levels=gmm_levels, alpha=alpha, colors="black")

            # Adjust display
            _ = ax.set_title("%s\n%s (%s)"%(ref_id,pos, ref_kmer))
            _ = ax.set_ylabel("log10 (Dwell Time)")
            _ = ax.set_xlabel("Median Intensity")
            _ = ax.set_xlim(xlim)
            _ = ax.set_ylim(ylim)
            _ = ax.legend()
            pl.tight_layout()

            return(fig, ax)

    def plot_volcano(self,
        ref_id:str,
        threshold:float=0.01,
        figsize:tuple=(30,10),
        palette:str="Set2",
        plot_style:str="ggplot",
        method:str="GMM_anova_pvalue"):
        """
        ###
        * ref_id
            Valid reference id name in the database
        * start
            Start coordinate
        * end
            End coordinate (included)
        * figsize
            Length and heigh of the output plot
        * palette
            Colormap. See https://matplotlib.org/users/colormaps.html, https://matplotlib.org/examples/color/named_colors.html
        * plot_style
            Matplotlib plotting style. See https://matplotlib.org/users/style_sheets.html
        * method
            Limit the pvalue methods shown in the plot. Either a list of methods or a regular expression as a string.
        * barplot
            plot p-value bars instead of lines
        """
        # Extract fasta and positions
        start, end = self.__get_positions(ref_id)

        try:
            ref_pos_dict = self.results.query('ref_id==@ref_id').set_index('pos').to_dict('index')
        except NameError:
            raise NanocomporeError("It looks like there's not results slot in SampCompDB")

        # Make a list with all methods available
        methods=list(self.results)
        if method not in self._metadata["pvalue_tests"]:
            raise NanocomporeError("Method %s is not in the results dataframe. Please chose one of %s "%(method, self._metadata["pvalue_tests"]))

        # Parse line position per position
        d = OrderedDict()


        rp=self[ref_id]
        for pos in range(start, end+1):
            # Collect results for position
            res_dict = OrderedDict()
            if pos in ref_pos_dict:
                for k,v in ref_pos_dict[pos].items():
                    if k == method:
                        if not np.isnan(v):
                            res_dict[k] = -np.log10(v)
                        else:
                            res_dict[k] = 0
                res_dict['ref_kmer'] = ref_pos_dict[pos]['ref_kmer']
                if 'GMM_model' in rp[pos]['txComp']:
                    res_dict['LOR'] = rp[pos]['txComp']['GMM_model'][1]
                else:
                    res_dict['LOR'] = 0
            d[pos] = res_dict


        # Cast collected results to dataframe
        df = pd.DataFrame.from_dict(d, orient="index")
        if df.empty:
            raise NanocomporeError("No data available for the selected interval")
        # Define plotting style
        with pl.style.context(plot_style):
            fig, ax = pl.subplots(figsize=figsize)
            _ = sns.scatterplot(x=df.LOR, y=df[method], palette=palette, ax=ax)
            for line in df.index.values:
                if df[method][line] > -np.log10(threshold):

                    ax.text(df.LOR[line], df[method][line], line, horizontalalignment='left', size='medium', color='black')
            _ = ax.axhline(y=-np.log10(threshold), color="grey", linestyle=":", label="pvalue={}".format(threshold))
            _ = ax.legend()
            _ = ax.set_ylabel("-log (pvalue)")
            _ = ax.set_xlabel("Log Odds Ratio")
            _ = ax.set_title(ref_id)
            pl.tight_layout()
            return(fig, ax)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nanocompore']

package_data = \
{'': ['*'], 'nanocompore': ['models/*']}

install_requires = \
['bedparse>=0.2,<0.3',
 'loguru>=0.5,<0.6',
 'matplotlib>=3.1,<3.2',
 'numpy>=1.19,<1.20',
 'pandas>=1,<2',
 'pyfaidx>=0.5,<0.6',
 'pyyaml>=5,<6',
 'scikit-learn>=0.23,<0.24',
 'scipy>=1.5,<1.6',
 'seaborn>=0,<1',
 'statsmodels>=0.12,<0.13',
 'tqdm>=4,<5']

entry_points = \
{'console_scripts': ['nanocompore = nanocompore.__main__:main']}

setup_kwargs = {
    'name': 'nanocompore',
    'version': '1.0.4',
    'description': 'Software package that identifies raw signal changes between two conditions from https://github.com/jts/nanopolish resquiggled dRNA-Seq data.',
    'long_description': '![Nanocompore](docs/pictures/Nanocompore_logo.png)\n\n---\n\n[![GitHub license](https://img.shields.io/github/license/tleonardi/nanocompore)](https://github.com/tleonardi/nanocompore/blob/master/LICENSE)\n[![Python](https://img.shields.io/badge/Python-%3E%3D3.6-yellow)](https://www.python.org/)\n[![BioRxiv](https://img.shields.io/badge/BioRxiv-10.1101%2F843136%20-red)](https://www.biorxiv.org/content/10.1101/843136v1.full)\n\n[![PyPI version](https://badge.fury.io/py/nanocompore.svg)](https://badge.fury.io/py/nanocompore)\n[![Downloads](https://pepy.tech/badge/nanocompore)](https://pepy.tech/project/nanocompore)\n[![Build Status](https://travis-ci.com/tleonardi/nanocompore.svg?token=2uTrW9fP9RypfMALjksc&branch=master)](https://travis-ci.com/tleonardi/nanocompore)\n\n---\n\n**Nanocompore identifies differences in ONT nanopore sequencing raw signal corresponding to RNA modifications by comparing 2 samples**\n\nNanocompore compares 2 ONT nanopore direct RNA sequencing datasets from different experimental conditions expected to have a significant impact on RNA modifications. It is recommended to have at least 2 replicates per condition. For example one can use a control condition with a significantly reduced number of modifications such as a cell line for which a modification writing enzyme was knocked-down or knocked-out. Alternatively, on a smaller scale transcripts of interests could be synthesized in-vitro.\n\n**Full documentation is available at http://nanocompore.rna.rocks**\n\n[![Nanocompore](docs/pictures/worflow.png)](http://nanocompore.rna.rocks)\n\n## Companion repositories\n\n* [NanoCompore_pipeline](https://github.com/tleonardi/nanocompore_pipeline): Nextflow pipeline to preprocess data for NanoCompore\n* [Nanocompore_analysis](https://github.com/tleonardi/nanocompore_paper_analyses): Analyses performed with Nanocompore for the BioRxiv preprint\n* [NanopolishComp](https://github.com/tleonardi/NanopolishComp): Collapse Nanopolish eventalign output per kmer, required before running NanoCompore\n\n## Main authors\n\n* Tommaso Leonardi - tom {at} tleo.io\n* Adrien Leger - aleg {at} ebi.ac.uk\n',
    'author': 'Tommaso Leonardi',
    'author_email': 'tom@itm6.xyz',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tleonardi/nanocompore',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-

#~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~#
# Std lib
from collections import *
import shelve
import multiprocessing as mp
import traceback
import datetime
import os

# Third party
from loguru import logger
import yaml
from tqdm import tqdm
import numpy as np
from pyfaidx import Fasta

# Local package
from nanocompore.common import *
from nanocompore.Whitelist import Whitelist
from nanocompore.TxComp import txCompare
from nanocompore.SampCompDB import SampCompDB
import nanocompore as pkg

# Disable multithreading for MKL and openBlas
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["MKL_THREADING_LAYER"] = "sequential"
os.environ["NUMEXPR_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"
os.environ['OPENBLAS_NUM_THREADS'] = '1'

#~~~~~~~~~~~~~~MAIN CLASS~~~~~~~~~~~~~~#
class SampComp(object):
    """ Init analysis and check args"""

    #~~~~~~~~~~~~~~FUNDAMENTAL METHODS~~~~~~~~~~~~~~#

    def __init__(self,
        eventalign_fn_dict:dict,
        fasta_fn:str,
        bed_fn:str = None,
        outpath:str = "results",
        outprefix:str = "out_",
        overwrite:bool = False,
        whitelist:Whitelist = None,
        comparison_methods:list = ["GMM", "KS"],
        logit:bool = True,
        anova:bool = False,
        allow_warnings:bool = False,
        sequence_context:int = 0,
        sequence_context_weights:str = "uniform",
        min_coverage:int = 30,
        min_ref_length:int = 100,
        downsample_high_coverage:int = 5000,
        max_invalid_kmers_freq:float = 0.1,
        select_ref_id:list = [],
        exclude_ref_id:list = [],
        nthreads:int = 3,
        progress:bool = False):

        """
        Initialise a `SampComp` object and generates a white list of references with sufficient coverage for subsequent analysis.
        The retuned object can then be called to start the analysis.
        * eventalign_fn_dict
            Multilevel dictionnary indicating the condition_label, sample_label and file name of the eventalign_collapse output.
            2 conditions are expected and at least 2 sample replicates per condition are highly recommended.
            One can also pass YAML file describing the samples instead.
            Example `d = {"S1": {"R1":"path1.tsv", "R2":"path2.tsv"}, "S2": {"R1":"path3.tsv", "R2":"path4.tsv"}}`
        * outpath
            Path to the output folder.
        * outprefix
            text outprefix for all the files generated by the function.
        * overwrite
            If the output directory already exists, the standard behaviour is to raise an error to prevent overwriting existing data
            This option ignore the error and overwrite data if they have the same outpath and outprefix.
        * fasta_fn
            Path to a fasta file corresponding to the reference used for read alignment.
        * bed_fn
            Path to a BED file containing the annotation of the transcriptome used as reference when mapping.
        * whitelist
            Whitelist object previously generated with nanocompore Whitelist. If not given, will be automatically generated.
        * comparison_methods
            Statistical method to compare the 2 samples (mann_whitney or MW, kolmogorov_smirnov or KS, t_test or TT, gaussian_mixture_model or GMM).
            This can be a list or a comma separated string. {MW,KS,TT,GMM}
        * logit
            Force logistic regression even if we have less than 2 replicates in any condition.
        * allow_warnings
            If True runtime warnings during the ANOVA tests don't raise an error.
        * sequence_context
            Extend statistical analysis to contigous adjacent base if available.
        * sequence_context_weights
            type of weights to used for combining p-values. {uniform,harmonic}
        * min_coverage
            minimal read coverage required in all sample.
        * min_ref_length
            minimal length of a reference transcript to be considered in the analysis
        * downsample_high_coverage
            For reference with higher coverage, downsample by randomly selecting reads.
        * max_invalid_kmers_freq
            maximum frequency of NNNNN, mismatching and missing kmers in reads.
        * select_ref_id
            if given, only reference ids in the list will be selected for the analysis.
        * exclude_ref_id
            if given, refid in the list will be excluded from the analysis.
        * nthreads
            Number of threads (two are used for reading and writing, all the others for parallel processing).
        * progress
            Display a progress bar during execution
        """
        logger.info("Checking and initialising SampComp")

        # Save init options in dict for later
        log_init_state(loc=locals())

        # If eventalign_fn_dict is not a dict try to load a YAML file instead
        if type(eventalign_fn_dict) == str:
            logger.debug("Parsing YAML file")
            if not access_file(eventalign_fn_dict):
                raise NanocomporeError("{} is not a valid file".format(eventalign_fn_dict))
            with open(eventalign_fn_dict, "r") as fp:
                eventalign_fn_dict = yaml.load(fp, Loader=yaml.SafeLoader)

        # Check eventalign_dict file paths and labels
        eventalign_fn_dict = self.__check_eventalign_fn_dict(eventalign_fn_dict)
        logger.debug(eventalign_fn_dict)

        # Check if fasta and bed files exist
        if not access_file(fasta_fn):
            raise NanocomporeError("{} is not a valid FASTA file".format(fasta_fn))
        if bed_fn and not access_file(bed_fn):
            raise NanocomporeError("{} is not a valid BED file".format(bed_fn))

        # Check threads number
        if nthreads < 3:
            raise NanocomporeError("The minimum number of threads is 3")

        # Parse comparison methods
        if comparison_methods:
            if type(comparison_methods) == str:
                comparison_methods = comparison_methods.split(",")
            for i, method in enumerate(comparison_methods):
                method = method.upper()
                if method in ["MANN_WHITNEY", "MW"]:
                    comparison_methods[i]="MW"
                elif method in ["KOLMOGOROV_SMIRNOV", "KS"]:
                    comparison_methods[i]="KS"
                elif method in ["T_TEST", "TT"]:
                    comparison_methods[i]="TT"
                elif method in ["GAUSSIAN_MIXTURE_MODEL", "GMM"]:
                    comparison_methods[i]="GMM"
                else:
                    raise NanocomporeError("Invalid comparison method {}".format(method))

        if not whitelist:
            whitelist = Whitelist(
                eventalign_fn_dict = eventalign_fn_dict,
                fasta_fn = fasta_fn,
                min_coverage = min_coverage,
                min_ref_length = min_ref_length,
                downsample_high_coverage = downsample_high_coverage,
                max_invalid_kmers_freq = max_invalid_kmers_freq,
                select_ref_id = select_ref_id,
                exclude_ref_id = exclude_ref_id)
        elif not isinstance(whitelist, Whitelist):
            raise NanocomporeError("Whitelist is not valid")

        # Set private args from whitelist args
        self.__min_coverage = whitelist._Whitelist__min_coverage
        self.__downsample_high_coverage = whitelist._Whitelist__downsample_high_coverage
        self.__max_invalid_kmers_freq = whitelist._Whitelist__max_invalid_kmers_freq

        # Save private args
        self.__eventalign_fn_dict = eventalign_fn_dict
        self.__db_fn = os.path.join(outpath, outprefix+"SampComp.db")
        self.__fasta_fn = fasta_fn
        self.__bed_fn = bed_fn
        self.__whitelist = whitelist
        self.__comparison_methods = comparison_methods
        self.__logit = logit
        self.__anova = anova
        self.__allow_warnings = allow_warnings
        self.__sequence_context = sequence_context
        self.__sequence_context_weights = sequence_context_weights
        self.__nthreads = nthreads - 2
        self.__progress = progress

        # Get number of samples
        n = 0
        for sample_dict in self.__eventalign_fn_dict.values():
            for sample_lab in sample_dict.keys():
                n+=1
        self.__n_samples = n

    def __call__(self):
        """
        Run the analysis
        """
        logger.info("Starting data processing")
        # Init Multiprocessing variables
        in_q = mp.Queue(maxsize = 100)
        out_q = mp.Queue(maxsize = 100)
        error_q = mp.Queue()

        # Define processes
        ps_list = []
        ps_list.append(mp.Process(target=self.__list_refid, args=(in_q, error_q)))
        for i in range(self.__nthreads):
            ps_list.append(mp.Process(target=self.__process_references, args=(in_q, out_q, error_q)))
        ps_list.append(mp.Process(target=self.__write_output, args=(out_q, error_q)))

        # Start processes and monitor error queue
        try:
            # Start all processes
            for ps in ps_list:
                ps.start()

            # Monitor error queue
            for tb in iter(error_q.get, None):
                logger.debug("Error caught from error_q")
                raise NanocomporeError(tb)
            logger.debug("Error queue was closed")

            # Soft processes and queues stopping
            logger.debug("Waiting for all processes to be joined")
            for ps in ps_list:
                ps.join()
            logger.debug("All processes joined successfully")

            logger.debug("Closing all queues")
            for q in (in_q, out_q, error_q):
                q.close()
            logger.debug("All queues were closed")

            # Return database wrapper object
            return SampCompDB(
                db_fn=self.__db_fn,
                fasta_fn=self.__fasta_fn,
                bed_fn=self.__bed_fn)

        # Catch error, kill all processed and reraise error
        except Exception as E:
            logger.error("An error occured. Killing all processes and closing queues\n")
            try:
                for ps in ps_list:
                    ps.terminate()
                for q in (in_q, out_q, error_q):
                    q.close()
            except:
                logger.error("An error occured while trying to kill processes\n")
            raise E

    #~~~~~~~~~~~~~~PRIVATE MULTIPROCESSING METHOD~~~~~~~~~~~~~~#
    def __list_refid(self, in_q, error_q):
        """Add valid refid from whitelist to input queue to dispatch the data among the workers"""
        n_tx = 0
        try:
            for ref_id, ref_dict in self.__whitelist:
                logger.debug("Adding {} to in_q".format(ref_id))
                in_q.put((ref_id, ref_dict))
                n_tx+=1

        # Manage exceptions and add error trackback to error queue
        except Exception:
            logger.debug("Error in Reader")
            error_q.put(traceback.format_exc())

        # Deal poison pills
        finally:
            for i in range (self.__nthreads):
                in_q.put(None)
            logger.debug("Parsed transcripts:{}".format(n_tx))

    def __process_references(self, in_q, out_q, error_q):
        """
        Consume ref_id, agregate intensity and dwell time at position level and
        perform statistical analyses to find significantly different regions
        """
        n_tx = n_reads = n_lines = 0
        try:
            logger.debug("Worker thread started")
            # Open all files for reading. File pointer are stored in a dict matching the ref_dict entries
            fp_dict = self.__eventalign_fn_open()

            # Process refid in input queue
            for ref_id, ref_dict in iter(in_q.get, None):
                logger.debug("Worker thread processing new item from in_q: {}".format(ref_id))
                # Create an empty dict for all positions first
                ref_pos_list = self.__make_ref_pos_list(ref_id)

                for cond_lab, sample_dict in ref_dict.items():
                    for sample_lab, read_list in sample_dict.items():
                        fp = fp_dict[cond_lab][sample_lab]

                        for read in read_list:

                            # Move to read, save read data chunk and reset file pointer
                            fp.seek(read["byte_offset"])
                            line_list = fp.read(read["byte_len"]).split("\n")
                            fp.seek(0)

                            # Check read_id ref_id concordance between index and data file
                            header = numeric_cast_list(line_list[0][1:].split("\t"))
                            if not header[0] == read["read_id"] or not header[1] == read["ref_id"]:
                                raise NanocomporeError("Index and data files are not matching:\n{}\n{}".format(header, read))

                            # Extract col names from second line
                            col_names = line_list[1].split("\t")
                            # Check that all required fields are present
                            if not all_values_in (["ref_pos", "ref_kmer", "median", "dwell_time"], col_names):
                                raise NanocomporeError("Required fields not found in the data file: {}".format(col_names))
                            # Verify if kmers events stats values are present or not
                            kmers_stats = all_values_in (["NNNNN_dwell_time", "mismatch_dwell_time"], col_names)

                            # Parse data files kmers per kmers
                            prev_pos = None
                            for line in line_list[2:]:
                                # Transform line to dict and cast str numbers to actual numbers
                                kmer = numeric_cast_dict(keys=col_names, values=line.split("\t"))
                                pos = kmer["ref_pos"]

                                # Check consistance between eventalign data and reference sequence
                                if kmer["ref_kmer"] != ref_pos_list[pos]["ref_kmer"]:
                                    ref_pos_list[pos]["ref_kmer"] = ref_pos_list[pos]["ref_kmer"]+"!!!!"
                                    #raise NanocomporeError ("Data reference kmer({}) doesn't correspond to the reference sequence ({})".format(ref_pos_list[pos]["ref_kmer"], kmer["ref_kmer"]))

                                # Fill dict with the current pos values
                                ref_pos_list[pos]["data"][cond_lab][sample_lab]["intensity"].append(kmer["median"])
                                ref_pos_list[pos]["data"][cond_lab][sample_lab]["dwell"].append(kmer["dwell_time"])
                                ref_pos_list[pos]["data"][cond_lab][sample_lab]["coverage"] += 1

                                if kmers_stats:
                                    # Fill in the missing positions
                                    if prev_pos and pos-prev_pos > 1:
                                        for missing_pos in range(prev_pos+1, pos):
                                            ref_pos_list[missing_pos]["data"][cond_lab][sample_lab]["kmers_stats"]["missing"] += 1
                                    # Also fill in with normalised position event stats
                                    n_valid = (kmer["dwell_time"]-(kmer["NNNNN_dwell_time"]+kmer["mismatch_dwell_time"])) / kmer["dwell_time"]
                                    n_NNNNN = kmer["NNNNN_dwell_time"] / kmer["dwell_time"]
                                    n_mismatching = kmer["mismatch_dwell_time"] / kmer["dwell_time"]
                                    ref_pos_list[pos]["data"][cond_lab][sample_lab]["kmers_stats"]["valid"] += n_valid
                                    ref_pos_list[pos]["data"][cond_lab][sample_lab]["kmers_stats"]["NNNNN"] += n_NNNNN
                                    ref_pos_list[pos]["data"][cond_lab][sample_lab]["kmers_stats"]["mismatching"] += n_mismatching
                                    # Save previous position
                                    prev_pos = pos

                                n_lines+=1
                            n_reads+=1

                logger.debug("Data for {} loaded.".format(ref_id))
                if self.__comparison_methods:
                    random_state=np.random.RandomState(seed=42)
                    ref_pos_list = txCompare(
                        ref_id=ref_id,
                        ref_pos_list=ref_pos_list,
                        methods=self.__comparison_methods,
                        sequence_context=self.__sequence_context,
                        sequence_context_weights=self.__sequence_context_weights,
                        min_coverage= self.__min_coverage,
                        allow_warnings=self.__allow_warnings,
                        logit=self.__logit,
                        anova=self.__anova,
                        random_state=random_state)

                # Add the current read details to queue
                logger.debug("Adding %s to out_q"%(ref_id))
                out_q.put((ref_id, ref_pos_list))
                n_tx+=1

        # Manage exceptions and add error trackback to error queue
        except Exception as e:
            logger.error("Error in Worker")
            error_q.put (NanocomporeError(traceback.format_exc()))

        # Deal poison pill and close file pointer
        finally:
            logger.debug("Processed Transcrits:{} Reads:{} Lines:{}".format(n_tx, n_reads, n_lines))
            logger.debug("Adding poison pill to out_q")
            self.__eventalign_fn_close(fp_dict)
            out_q.put(None)

    def __write_output(self, out_q, error_q):
        # Get results out of the out queue and write in shelve
        pvalue_tests = set()
        ref_id_list = []
        n_tx = n_pos = 0
        try:
            with shelve.open(self.__db_fn, flag='n') as db, tqdm(total=len(self.__whitelist), unit=" Processed References", disable= not self.__progress) as pbar:
                # Iterate over the counter queue and process items until all poison pills are found
                for _ in range(self.__nthreads):
                    for ref_id, ref_pos_list in iter(out_q.get, None):
                        ref_id_list.append(ref_id)
                        logger.debug("Writer thread writing %s"%ref_id)
                        # Get pvalue fields available in analysed data before
                        for pos_dict in ref_pos_list:
                            if 'txComp' in pos_dict:
                                for res in pos_dict['txComp'].keys():
                                    if "pvalue" in res:
                                        n_pos+=1
                                        pvalue_tests.add(res)
                        # Write results in a shelve db
                        db [ref_id] = ref_pos_list
                        pbar.update(1)
                        n_tx+=1

                # Write list of refid
                db["__ref_id_list"] = ref_id_list

                # Write metadata
                db["__metadata"] = {
                    "package_name": pkg.__version__,
                    "package_version": pkg.__name__,
                    "timestamp": str(datetime.datetime.now()),
                    "comparison_methods": self.__comparison_methods,
                    "pvalue_tests": sorted(list(pvalue_tests)),
                    "sequence_context": self.__sequence_context,
                    "min_coverage": self.__min_coverage,
                    "n_samples": self.__n_samples}

        # Manage exceptions and add error trackback to error queue
        except Exception:
            logger.error("Error in Writer")
            error_q.put(traceback.format_exc())

        finally:
            logger.debug("Written Transcripts:{} Valid positions:{}".format(n_tx, n_pos))
            logger.info ("All Done. Transcripts processed: {}".format(n_tx))
            # Kill error queue with poison pill
            error_q.put(None)

    #~~~~~~~~~~~~~~PRIVATE HELPER METHODS~~~~~~~~~~~~~~#
    def __check_eventalign_fn_dict(self, d):
        """"""
        # Check that the number of condition is 2 and raise a warning if there are less than 2 replicates per conditions
        if len(d) != 2:
            raise NanocomporeError("2 conditions are expected. Found {}".format(len(d)))
        for cond_lab, sample_dict in d.items():
            if len(sample_dict) == 1:
                logger.info("Only 1 replicate found for condition {}".format(cond_lab))
                logger.info("This is not recommended. The statistics will be calculated with the logit method")

        # Test if files are accessible and verify that there are no duplicated replicate labels
        duplicated_lab = False
        rep_lab_list = []
        rep_fn_list = []
        for cond_lab, sd in d.items():
            for rep_lab, fn in sd.items():
                if not access_file(fn):
                    raise NanocomporeError("Cannot access eventalign file: {}".format(fn))
                if fn in rep_fn_list:
                    raise NanocomporeError("Duplicated eventalign file detected: {}".format(fn))
                if rep_lab in rep_lab_list:
                    duplicated_lab = True
                rep_lab_list.append(rep_lab)
                rep_fn_list.append(fn)
        if not duplicated_lab:
            return d

        # If duplicated replicate labels found, prefix labels with condition name
        else:
            logger.debug("Found duplicated labels in the replicate names. Prefixing with condition name")
            d_clean = OrderedDict()
            for cond_lab, sd in d.items():
                d_clean[cond_lab] = OrderedDict()
                for rep_lab, fn in sd.items():
                    d_clean[cond_lab]["{}_{}".format(cond_lab, rep_lab)] = fn
            return d_clean

    def __eventalign_fn_open(self):
        """"""
        fp_dict = OrderedDict()
        for cond_lab, sample_dict in self.__eventalign_fn_dict.items():
            fp_dict[cond_lab] = OrderedDict()
            for sample_lab, fn in sample_dict.items():
                fp_dict[cond_lab][sample_lab] = open(fn, "r")
        return fp_dict

    def __eventalign_fn_close(self, fp_dict):
        """"""
        for sample_dict in fp_dict.values():
            for fp in sample_dict.values():
                fp.close()

    def __make_ref_pos_list(self, ref_id):
        """"""
        ref_pos_list = []
        with Fasta(self.__fasta_fn) as fasta:
            ref_fasta = fasta [ref_id]
            ref_len = len(ref_fasta)
            ref_seq = str(ref_fasta)

            for pos in range(ref_len-4):
                pos_dict = OrderedDict()
                pos_dict["ref_kmer"] = ref_seq[pos:pos+5]
                pos_dict["data"] = OrderedDict()
                for cond_lab, s_dict in self.__eventalign_fn_dict.items():
                    pos_dict["data"][cond_lab] = OrderedDict()
                    for sample_lab in s_dict.keys():

                        pos_dict["data"][cond_lab][sample_lab] = {
                            "intensity":[],
                            "dwell":[],
                            "coverage":0,
                            "kmers_stats":{"missing":0,"valid":0,"NNNNN":0,"mismatching":0}}
                ref_pos_list.append(pos_dict)
        return ref_pos_list
